from gcandle.core.buy_strategy.BuyStrategy import BuyStrategy
from gcandle.core.indicator_service.master_indicator_service import MasterIndicatorService
from gcandle.utils.date_time_utils import Date
DAYS_BEFORE_TO_SCAN = 15


def to_jzsec_code(c):
    if c.startswith('00') or c.startswith('300'):
        return 'SZSE.' + c
    else:
        return 'SHSE.' + c


class DayBarCandidatesScanner:
    def __init__(self, name, master_service: MasterIndicatorService, buy_strategy: BuyStrategy):
        self.name = name
        self.master_service = master_service
        self.buy_strategy = buy_strategy
        self.buy_ratio = None

    def set_buy_ratio(self, ratio):
        self.buy_ratio = ratio

    def scan(self, argv):
        if len(argv) > 1:
            codes = self.master_service.read_all_codes()
            end = argv[1]
            start = Date(end).get_before(DAYS_BEFORE_TO_SCAN).as_str()
            print('## Scan {} candidates with backtest data from {} to {}'.format(self.name, start, end))
            data = self.master_service.read_with_slave_for_backtest(codes, start, end=end)
        else:
            print('## Scan {} candidates for trade'.format(self.name))
            data = self.master_service.read_with_slave_for_trade()
        self.__do_scan(data)

    def __do_scan(self, data):
        next_trade_day = str(data.index.levels[0].max())[0:10]
        print("Target trade date: {}".format(next_trade_day))
        buy_strategy = self.buy_strategy
        data = buy_strategy.calc_and_forward_fill_buy_filter(data, True)
        buy_strategy.update_buy_prices(data)

        data = data.reset_index()
        buy_filter = buy_strategy.get_buy_filter_before_buy_day(data)
        day_filter = (data.date == next_trade_day)
        data.loc[buy_filter & day_filter, 'to_buy'] = data.bpr

        data['target_chg'] = data['to_buy'] / data.pc
        data = data.loc[data.target_chg > 0]

        if data.shape[0] > 0:
            self.__save_candidates(data, next_trade_day)
        else:
            print('No candidates on {}'.format(next_trade_day))

    def __smart_buy_ratio(self, n):
        r = 0.35
        if n > 20:
            r = 0.25
        elif n > 40:
            r = 0.2
        return r

    def __save_candidates(self, data, next_day):

        if data.shape[0] > 0:
            print('{} candidates identified on {} for {}'.format(data.shape[0], next_day, self.name))
            buy_ratio = self.buy_ratio
            if buy_ratio is None:
                buy_ratio = self.__smart_buy_ratio(len(data))

            target_list = [[to_jzsec_code(r.code), round(r.to_buy, 2),
                            buy_ratio,  round(r.target_chg, 4)]
                           for r in data.itertuples(index=False)]
            fname = '{}_juejin_list.txt'.format(self.name)
            f = open(fname, 'w')
            print('g_today = \'{}\''.format(next_day), file=f)
            print('g_targets_to_buy = [', file=f)
            for r in target_list:
                print(r, ',', file=f)
            print(']', file=f)
            f.close()
        else:
            print('No candidates on {}'.format(next_day))


if __name__ == '__main__':
    print("Common Day Bar Candidates Scanner class")



import autosklearn.classification
import sklearn.model_selection
import sklearn.datasets
import sklearn.metrics

X, y = sklearn.datasets.load_digits(return_X_y=True)
print('loaded')
X_train, X_test, y_train, y_test = \
        sklearn.model_selection.train_test_split(X, y, random_state=1)
print('splitted')
automl = autosklearn.classification.AutoSklearnClassifier(include_estimators=['liblinear_svc'])
automl.fit(X_train, y_train)
print('fited')
y_hat = automl.predict(X_test)
print("Accuracy score", sklearn.metrics.accuracy_score(y_test, y_hat))


from gcandle.objects.basic_objects import DB_CLIENT
from gcandle.utils.multi_process_pool import execute_tasks

BASIC_DATA_COLUMNS = ['open', 'close', 'low', 'high', 'amount']

START_DATE = '2001-01-01'


class SecurityDataVerifier:
    def __init__(self):
        pass

    def compare_basic_data_of_repos(self, code, start, end, repo1, repo2, cols):
        df1 = DB_CLIENT.read_data_by_codes_and_date(repo1, [code], start, end)
        if df1 is None:
            print("no data from {} for {}".format(repo1, code))
            return None
        df1 = df1[cols]
        df2 = DB_CLIENT.read_data_by_codes_and_date(repo2, [code], start, end)
        if df2 is None:
            print("no data from {} for {}".format(repo2, code))
            return None
        df2 = df2[cols]
        res = self.do_compare(df1, df2, cols)
        if res is None:
            print("Something is wrong")
            return
        l, r, d = res
        if len(l) > 0:
            print("### {} Left only ###: \n".format(code), l)
        if len(r) > 0:
            print('### {} Right only ###: \n'.format(code), r)
        if len(d) > 0:
            print('### {} Not matched ###: \n'.format(code), d)
        if len(l) == 0 and len(r) == 0 and len(d) == 0:
            print("{} All matched".format(code))

    def do_compare(self, df1, df2, cols):
        joined = df1.join(df2, how='outer', rsuffix='_r')
        col = cols[0]
        left_only = joined[joined[col + '_r'].isna()]
        right_only = joined[joined[col].isna()]
        both_have = joined.dropna()
        if len(both_have) < 1:
            print('Warning: no data that exists in both dataframes')
            return left_only, right_only, both_have
        else:
            value_not_match = both_have[col] != both_have[col + '_r']
            for col in cols[1:]:
                value_not_match |= both_have[col] != both_have[col + '_r']
            return left_only, right_only, both_have[value_not_match]


def verify_qfq_data():
    repo1 = 'stock_day_qfq'
    repo2 = 'stock_day_qfq_new_stock_day_with_old_xdxr'
    # codes = ['600158']
    execute_tasks(codes, verifier.compare_basic_data_of_repos, start, end, repo1, repo2, BASIC_DATA_COLUMNS)


def verify_xian_backtest():
    repo1 = 'xian_backtest'
    repo2 = 'xian_stock_day_long_fp'
    # codes = ['600158']
    execute_tasks(codes, verifier.compare_basic_data_of_repos, start, end, repo1, repo2, ['open', 'close'])


def verify_fp_marks_data():
    for fpd in [20, 50, 120]:
        repo1 = 'fp_marks_' + str(fpd)
        repo2 = 'indicator_fp_marks_' + str(fpd) + '_day'
        cols = ['fp_type', 'fp_around_' + str(fpd)]
        execute_tasks(codes, verifier.compare_basic_data_of_repos, start, end, repo1, repo2, cols)

def verify_chan_zs():
    repo1 = 'indicator_chan_current_zs_1min'
    repo2 = 'indicator_chan_current_zs_1min_copy'
    # codes = ['600158']
    execute_tasks(codes, verifier.compare_basic_data_of_repos, start, end, repo1, repo2, ['xd_time', 'zs_time', 'zs_low', 'zs_high'])


def verify_chan_fp():
    repo1 = 'indicator_chan_current_fp_1min'
    repo2 = 'indicator_chan_current_fp_1min_copy'
    # codes = ['600158']
    execute_tasks(codes, verifier.compare_basic_data_of_repos, start, end, repo1, repo2, ['pFpTime', 'pFpType', 'pFpPrice'])

if __name__ == '__main__':
    from gcandle.data.SecurityDataService import SecurityDataService
    import random
    dataService = SecurityDataService()
    verifier = SecurityDataVerifier()
    start = '2019-01-01'
    end = '2020-09-31'
    codes = dataService.read_security_codes()
    codes = random.sample(codes, 300)
    codes = ['000001']
    verify_chan_fp()
    # verify_fp_marks_data()




from gcandle.core.indicator_service.master_indicator_service import MasterIndicatorService
import pandas as pd

show_cols = ['low', 'high', 'profit']


def fill_future_prices(data):
    cols = ['high', 'close']
    def _fill(data):
        for n in [1, 2]:
            for col in cols:
                data[col + 'A' + str(n)] = data[col].shift(-1 * n)
        return data
    data = data.groupby(level=1).apply(_fill).sort_index(level=0)
    return data


class StrategyAnalyzer:
    BUY_PR_KEY='bpr'
    SELL_PR_KEY='spr'
    PROFIT_KEY='profit'

    def __init__(self, master_service: MasterIndicatorService, start: str, end: str, sample=1.0):
        self.master_service = master_service
        self.sample = sample
        self.data = master_service.read_with_slave_for_backtest(start, end, ratio=sample)

    def fill_master_filter(self, data):
        #to be implemented by concrete analyzer
        return None

    def calc_buy_and_sell_price(self, data):
        #to be implemented by concrete analyzer
        return None

    def get_buy_day_filter(self, data):
        #to be implemented by concrete analyzer
        return None

    def calc_profit(self):
        data = fill_future_prices(self.data)
        data = self.fill_master_filter(data)
        self.calc_buy_and_sell_price(data)
        buy_filter = self.get_buy_day_filter(data)
        data = data.loc[buy_filter]
        data[StrategyAnalyzer.PROFIT_KEY] = \
            data[StrategyAnalyzer.SELL_PR_KEY] / data[StrategyAnalyzer.BUY_PR_KEY]
        self.data = data.copy()
        return self

    def get_stats(self):
        bins = [0, 1, 100]
        n = len(bins)
        data = self.data
        ntotal = len(self.data)
        cols = ['gratio', 'pr0', 'pr1', 'glen', 'gmean', 'gmedian', 'gprod']
        res = []
        for i in range(0, n - 1):
            gf = (data.profit >= bins[i]) & (data.profit < bins[i + 1])
            g = data.loc[gf]
            glen = len(g)
            gratio = round(glen / ntotal, 2)
            gmean = round(g.profit.mean(), 3)
            gmedian = round(g.profit.median(), 3)
            gprod = g.profit.product()
            res.append([gratio, bins[i], bins[i + 1], glen, gmean, gmedian, gprod])
        res.append([1.0, 0, 100, len(data), round(data.profit.mean(), 3),
                    round(data.profit.median(), 3), data.profit.product()])
        return pd.DataFrame(res, columns=cols)

    def show_losses(self, pr=0.94, extra_cols=None):
        data = self.data
        loss = data.loc[data.profit < pr]
        print('len: ', len(loss))
        cols = show_cols.copy()
        if extra_cols is not None:
            cols += extra_cols
        # loss[cols].to_csv('lian_yang_loss.csv')
        return loss[cols].tail(20)

    def show_win(self, data, pr=1.05, extra_cols=None):
        win = data.loc[data.profit > pr]
        print('len: ', len(win))
        cols = show_cols.copy()
        if extra_cols is not None:
            cols += extra_cols
        #     print(win[cols])
        # win[cols].to_csv('lian_yang_win.csv')
        return win[cols].tail(20)

    def get_filtered(self, data, f, extra_cols=None):
        win = data.loc[f]
        print('len: ', len(win))
        cols = show_cols.copy()
        print(cols)
        print('extra, ', extra_cols)
        if extra_cols is not None:
            cols += extra_cols
        if len(win) < 200:
            print(win[cols])
        win[cols].to_csv('lian_yang_filtered.csv')
        return win



import gcandle.utils.dummy_appender as dummy_appender
import gcandle.utils.date_time_utils as date_time_utils
import gcandle.utils.multi_process_pool as executor
from gcandle.objects.data_service_objects import SECURITY_DATA_READ_SERVICE
from gcandle.objects.basic_objects import DB_CLIENT


class MasterIndicatorDB:
    MAX_CALENDAR_DAYS_TO_KEEP = 15
    DEFAULT_DUMMY_ROWS = 10
    BACKTEST_POSTFIX = '_backtest'
    TRADE_POSTFIX = '_dummy_for_trade'

    def __init__(self, name, indicator_func=None, dummy_rows=DEFAULT_DUMMY_ROWS):
        self.name = name
        self.indicator_func = indicator_func
        self.dummy_rows = dummy_rows

    def set_dummy_rows(self, n):
        self.dummy_rows = n

    def set_indicator_func(self, func):
        self.indicator_func = func

    def repo_name_for_backtest(self):
        return self.name + MasterIndicatorDB.BACKTEST_POSTFIX

    def repo_name_for_trade(self):
        return self.name + MasterIndicatorDB.TRADE_POSTFIX

    def read_train_for_backtest(self, codes, start, end):
        return DB_CLIENT.read_data_by_codes_and_date(
            self.repo_name_for_backtest(),
            codes, start, end=end, extra_filter={'train': True}
        )

    def read_test_for_backtest(self, codes, start, end):
        return DB_CLIENT.read_data_by_codes_and_date(
            self.repo_name_for_backtest(),
            codes, start, end=end, extra_filter={'train': False}
        )

    def read_for_backtest(self, codes, start, end):
        return DB_CLIENT.read_data_by_codes_and_date(
            self.repo_name_for_backtest(),
            codes, start, end=end
        )

    def read_for_trade(self):
        return DB_CLIENT.read_all_data(
            self.repo_name_for_trade()
        )

    def read_all_codes(self):
        return DB_CLIENT.read_codes_from_db(self.repo_name_for_backtest())

    def update_for_single_code(self, code, start, end):
        data = SECURITY_DATA_READ_SERVICE.read_qfq_security_data_for_single(code, start, end)
        if data is not None and len(data) > 1:
            data = dummy_appender.append_dummy_rows(data, 1)
            data = self.indicator_func(data)
            if data is None or data.shape[0] < 1:
                pass
            else:
                df_without_dummy = dummy_appender.exclude_dummy_rows(data)
                DB_CLIENT.update_data_append_newer(df_without_dummy, self.repo_name_for_backtest())

                df_with_dummy_for_trade = data.tail(self.dummy_rows)
                min_start_date = date_time_utils.Date.from_str(end). \
                    get_before(days=MasterIndicatorDB.MAX_CALENDAR_DAYS_TO_KEEP)
                df_with_dummy_for_trade = df_with_dummy_for_trade.reset_index(level=0)
                df_with_dummy_for_trade = df_with_dummy_for_trade.loc[df_with_dummy_for_trade.date > min_start_date.as_str()]
                DB_CLIENT.replace_data_by_codes(df_with_dummy_for_trade, self.repo_name_for_trade())

    def update_all_codes(self, days_to_update=60):
        code_list = SECURITY_DATA_READ_SERVICE.read_security_codes()
        end = date_time_utils.Date()
        start = end.get_before(days=int(days_to_update * 1.5))
        executor.execute_tasks(code_list, self.update_for_single_code, start.as_str(), end.as_str())

    def recreate_all_codes(self, start_date, end):
        DB_CLIENT.remove_all_from(self.repo_name_for_backtest())
        DB_CLIENT.remove_all_from(self.repo_name_for_trade())
        DB_CLIENT.create_index_for_collection_with_date_and_code(self.repo_name_for_backtest())
        DB_CLIENT.create_index_for_collection_with_date_and_code(self.repo_name_for_trade())

        code_list = SECURITY_DATA_READ_SERVICE.read_security_codes()
        start = date_time_utils.Date.from_str(start_date)
        end = date_time_utils.Date.from_str(end)
        executor.execute_tasks(code_list, self.update_for_single_code, start.as_str(), end.as_str())




import copy
import datetime

import numpy as np
import pandas as pd

from gcandle.core.common_types import ORDER_DIRECTION
from gcandle.core.backtest.AccountPositionStock import AccountPositionStock


def up_limit_ultimate(data):
    return (data.open > data.pc) and (data.open == data.close == data.high == data.low)


def down_limit_ultimate(data):
    return (data.open < data.pc) and (data.open == data.close == data.high == data.low)


class BacktestAccount:
    def __init__(
            self,
            init_cash=1000000,
    ):
        self._history_headers = [
            'datetime',  # 日期/时间
            'code',  # 品种
            'price',  # 成交价
            'amount',  # 成交数量(股票 股数  期货 手数)
            'cash',  # 现金
            'commission',  # 手续费
            'tax',  # 税
            'direction'  # 方向
        ]
        self.init_cash = init_cash
        self.init_hold = {}
        self.current_hold = {}
        self.history = []
        self.hold = {}
        self.fee_ratio = 0.00015
        self.tax_ratio = 0.001
        self._cash_table = []
        self.cash_available = self.init_cash
        self.sell_available = copy.deepcopy(self.init_hold)
        self.buy_available = copy.deepcopy(self.init_hold)
        self.history = []
        self.static_balance = {
            'static_assets': [],
            'cash': [],
            'frozen': [],
            'hold': [],
            'date': []
        }
        self.today_trade = {'last': [], 'current': []}
        self.today_orders = {'last': [], 'current': []}
        self.positions = AccountPositionStock()

    def __repr__(self):
        return 'account'

    @property
    def message(self):
        'the standard message which can be transfer'
        return {
            'source': 'account',
            'init_cash': self.init_cash,
            'init_hold': self.init_hold,
            'cash': self.cash_available,
            'history': self.history,
        }

    @property
    def start_date(self):
        return self._cash_table[0][0]

    @property
    def end_date(self):
        return self._cash_table[-1][0]

    def start(self, today):
        self._cash_table.append([today, self.init_cash])

    def take_snapshot(self, today):
        self._cash_table.append([today, self.cash_available])

    def settle(self, data):
        self.sell_available = copy.deepcopy(self.hold)
        self.cash_available = self.cash_available
        date = data.date[0]
        self.take_snapshot(date)
        self.positions.close_trade(data)

    def get_daily_assets(self):
        daily_pos = self.positions.get_daily_positions()
        daily_pos = np.array(daily_pos).T
        dp = pd.DataFrame(data={'date': daily_pos[0], 'se_value': daily_pos[3]})
        daily_se_value = dp.groupby(['date'], as_index=True).agg({'se_value': sum})
        daily_cash = self.daily_cash
        daily_assets = pd.concat([daily_se_value, daily_cash], join='outer', axis=1)
        daily_assets.se_value.fillna(0, inplace=True)
        daily_assets.cash.fillna(method='ffill', inplace=True)
        daily_assets['asset'] = daily_assets['se_value'] + daily_assets['cash']
        daily_assets = daily_assets.round(0)
        return daily_assets

    def get_active_position_days(self, code):
        return self.positions.get_active_position_days(code)

    def get_all_positions(self):
        return self.positions.get_all_positions()

    def win_rate(self):
        return self.positions.win_rate()

    def deal(self, data, price, amount, direction, money, fee, tax=0):
        cash_after_deal = self.cash_available - money * direction
        if cash_after_deal < 0:
            print('Invalid deal!!!! Cash less than 0')
            return
        today = data.date
        self.history.append(
            [
                str(today),
                data.code,
                price,
                amount*direction,
                self.cash_available,
                fee,
                tax,
                direction
            ]
        )
        self.cash_available = cash_after_deal
        hold = self.hold
        if data.code in hold:
            orig_hold = hold[data.code]
            orig_hold['amount'] += amount * direction
            if orig_hold['amount'] <= 0:
                # print('{} sold out.'.format(data.code))
                del hold[data.code]
        else:
            hold[data.code] = {'date': today, 'amount': amount, 'price': price}

    def validate_amount(self, amount, data, buy=True):
        # 1手100股
        max_ratio = 0.05 * 100 * 1000
        if amount > (data.vol * max_ratio):
            print('{} {} order too much, fulfill ratio = {}'.format(
                str(data.date)[0:10], data.code, round(data.vol * max_ratio / amount, 2)))
            amount = int(data.vol * max_ratio)
        if buy:
            amount = amount - amount % 100
        return amount

    def validate_price(self, order_price, data):
        return data.high >= order_price >= data.low

    def exclude_yi_zi_ban_up(self, data):
        if 'pc' in data.columns:
            f = data.close / data.pc > 1
            f &= (data.low == data.high)
            return data.loc[~f]
        else:
            return data

    def batch_buy(self, data, price_func, max_ratio=1.0):
        cash_today = self.cash_available
        data = self.exclude_yi_zi_ban_up(data)

        ratio = round(1.0/len(data), 2) - 0.01

        if ratio > max_ratio:
            ratio = max_ratio
        for row in data.itertuples():
            price = price_func(row)
            amount = (cash_today * ratio) / (price * (1+self.fee_ratio + self.tax_ratio))
            self.order(row, ORDER_DIRECTION.BUY, amount=amount, price=price)

    def order(self, data, towards, ratio=0, amount=0, price=0, price_func=None):
        cash = self.cash_available
        if price is None:
            if price_func is None:
                print("Invalid parameters, both price and price_func are None!")
                return False
            else:
                price = price_func(data)
        order_dir = ''
        if towards == ORDER_DIRECTION.BUY:
            order_dir = 'BUY'
        else:
            order_dir = 'SELL'
        #for debug
        # print('{} {} {}  - {}'.format(data.date, order_dir, data.code, cash))
        if not self.validate_price(price, data):
            print(order_dir + ' Order price {} out of range [{}, {}]'.format(price, data.low, data.high))
            return False
        if towards == ORDER_DIRECTION.BUY:
            if up_limit_ultimate(data):
                print(data.date, data.code, '涨停一字，无法买入')
                return False
            # price = (data.low + data.high) * 0.5
            # print('Buy {} {} '.format(data.code, price))
            if amount == 0:
                if ratio == 0:
                    print("Invalid parameters, both amount and ratio are 0.")
                    return False
                else:
                    to_pay = min(cash * ratio, cash)
                    amount = to_pay / price * (1 + self.fee_ratio + self.tax_ratio)
            amount = self.validate_amount(amount, data)
            if amount <= 0:
                # print('Failed to deal due to cash or vol')
                return False
            fee = self.fee_ratio * price * amount
            fee = max(fee, 5)
            tax = 0
            to_pay = amount * price + fee
            if cash < to_pay:
                print('Not enough money or invalid money calc', data.date, data.code, price, amount)
                return False
            self.positions.open_position(price, amount, data)
            self.deal(data=data, price=price, amount=amount, direction=1, money=to_pay, fee=fee, tax=tax)
        else:
            if down_limit_ultimate(data):
                print(data.date, data.code, '跌停一字，无法卖出。')
                #价格已检验过，此处只是假设再跌20%。
                # price = price * 0.8
                return False
            if data.code not in self.hold:
                print('Invalid order, no sellable stocks for ', data.code)
                return False
            sellable = self.hold[data.code]
            amount = sellable['amount'] * ratio
            amount = self.validate_amount(amount, data)
            if amount <= 0:
                # print('Failed to deal due to vol')
                return False
            # print('Sell {} {} '.format(data.code, price))
            fee = price * amount * self.fee_ratio
            tax = price * amount * self.tax_ratio
            money = price * amount - fee - tax
            # print('sell {} deal OK'.format(data.code))
            self.positions.close_position(price, data)
            self.deal(data=data, price=price, amount=amount, direction=-1, money=money, fee=fee, tax=tax)
        return True

    @property
    def history_table(self):
        if len(self.history) > 0:
            lens = len(self.history[0])
        else:
            lens = len(self._history_headers)

        return pd.DataFrame(
            data=self.history,
            columns=self._history_headers[:lens]
        ).sort_index()

    @property
    def daily_cash(self):
        _cash = pd.DataFrame(
            data=self._cash_table, columns=['date', 'cash']).\
            set_index('date', drop=True)
        return _cash

    @property
    def profit_by_stock(self):
        h = self.history_table

        def _calc_prof_sum(x):
            r = pd.Series({'profit_sum': x.amount.sum() * x.iloc[-1].price - (x['amount'] * x.price).sum()})
            return r

        if h.shape[0] > 0:
            return h.groupby('code').apply(_calc_prof_sum)
        else:
            return None

    @property
    def profit_rate_by_stock(self):
        h = self.history_table

        def _calc_prof_rate(x):
            x.loc[x.direction == -1, 'bprice'] = x.price.shift(1)
            x['order_rate'] = x.price / x.bprice
            x.order_rate.fillna(1, inplace=True)
            x['profit_rate'] = x.order_rate.prod()
            return pd.Series({'profit_rate': x.profit_rate.iloc[-1]})

        if h.shape[0] > 0:
            return h.groupby('code').apply(_calc_prof_rate)
        else:
            return None

    def max_profitable_stocks(self, n):
        profits = self.profit_rate_by_stock
        if profits is not None:
            if n > 0:
                return profits.sort_values(by=['profit_rate']).tail(n)
            else:
                return profits.sort_values(by=['profit_rate']).head(abs(n))
        else:
            return None



from multiprocessing import Pool
from gcandle.objects.basic_objects import DB_CLIENT
from time import time

DEFAULT_POOL_SIZE = 8


def pool_init():
    DB_CLIENT.init()


def foo(*args):
    print('foo: ', len(args))
    print(args)
    print(*args)


def execute_tasks(code_list, task_func, *args):
    n_start = 0
    n_end = len(code_list)
    code_list = code_list[n_start:n_end]

    t_start = time()
    pool = Pool(DEFAULT_POOL_SIZE, initializer=pool_init)
    if len(args) > 0:
        results = [pool.apply_async(task_func, args=(c, *args)) for c in code_list]
    else:
        # Fucking python feature: if a tuple literal is created with one single str,
        # the str will be splitted as chars
        results = [ pool.apply_async(task_func, args=(c, )) for c in code_list ]
    count = n_start
    total = len(code_list) + n_start
    for r in results:
        insert_result = r.get()
        count += 1
        if count % 300 == 0:
            print('Updating progress {:.2f}%, {}'.format(float(count / total) * 100.0, count))
        if insert_result is not None:
            if insert_result.acknowledged:
                pass
            else:
                print('Failed to insert results')

    t_end = time()
    print('Time used: ', t_end - t_start)




from gcandle.data.se_types import SeType, SeFreq
from gcandle.objects.data_service_objects import CSV_SECURITY_DATA_SERVICE, SECURITY_DATA_READ_SERVICE
from gcandle.data.SecurityDataUpdateMonitor import SecurityDataUpdateMonitor
from gcandle.utils.date_time_utils import Date
import time


def fetch_and_update_min_bars_by_csv():
    dataService = CSV_SECURITY_DATA_SERVICE
    monitor = SecurityDataUpdateMonitor()
    codes = SECURITY_DATA_READ_SERVICE.read_security_codes()
    # codes = ['000001']
    a = Date()
    start = '2018-01-01'
    end = '2018-12-31'
    # monitor.add_monitor_repo("stock_1min")
    # monitor.add_monitor_repo("stock_5min")
    # monitor.add_monitor_repo("stock_30min")
    # monitor.collect_baseline()
    # codes = ['000001', '300343'] + codes[:100]
    dataService.update_bars_by_dates(start, end, typ=SeType.Stock, freq=SeFreq.ONE_MIN, codes=codes)
    # dataService.refetch_and_save_bars(typ=SeType.Stock, freq=SeFreq.FIVE_MIN, codes=codes)
    # dataService.update_bars(typ=SeType.Stock, freq=SeFreq.THIRTY_MIN, codes=None)
    # monitor.collect_info_after_update(a)
    b = Date()
    print("time used: {} seconds".format(b.delta_to(a).seconds))
    # monitor.report_changes()


if __name__ == '__main__':
    fetch_and_update_min_bars_by_csv()


import gcandle.utils.date_time_utils as date_time_utils
import pandas as pd

dummy_key = 'dummy_row'
dummy_dict = {
    'open': 0,
    'close': 0,
    'low': 0,
    'high': 0,
    'volume': 0,
    'amount': 0,
    dummy_key: True
}


def get_dummy_day_row(date, code):
    index = pd.MultiIndex.from_arrays([[date], [code]], names=('date', 'code'))
    return pd.DataFrame(data=dummy_dict, index=index)


def append_dummy_rows(data, n_rows=1):
    last_date = data.index.levels[0][-1]
    code = data.index.levels[1][0]
    next_day = date_time_utils.Date.from_datetime(last_date)
    for n in range(n_rows):
        next_day = next_day.next_trade_day()
        dummy_row = get_dummy_day_row(next_day.as_datetime(), code)
        data = data.append(dummy_row, sort=True)
    return data


def get_dummy_rows(data):
    if dummy_key in data.columns:
        return data.loc[data[dummy_key] == True]
    else:
        return None


def exclude_dummy_rows(data):
    if dummy_key in data.columns:
        data = data.loc[data[dummy_key].isna() | (data[dummy_key] == False)]
        data = data.drop(columns=[dummy_key])
    return data



from gcandle.objects.basic_objects import DB_CLIENT
from gcandle.objects.data_service_objects import SECURITY_DATA_READ_SERVICE
from gcandle.utils.date_time_utils import Date

START_DATE = '2001-01-01'
COLLECTIONS_TO_CHECK = [
    'stock_day',
    'index_day',
    'stock_day_qfq',
    'security_list',
    'security_xdxr',
    'xian_backtest',
    'xian_dummy_for_trade',
    'da_lian_yang_backtest',
    'da_lian_yang_dummy_for_trade',
    'indicator_fp_marks_20_day',
    'indicator_fp_marks_50_day',
    'indicator_fp_marks_120_day',
    'fp_features_20_xian',
    'fp_features_50_xian',
    'fp_features_120_xian',
    'fp_features_20_da_lian_yang',
    'fp_features_50_da_lian_yang',
    'fp_features_120_da_lian_yang',
]


class SecurityDataUpdateMonitor:
    def __init__(self):
        self.repos_to_monitor = []
        self.baseline = {}
        self.info_after = {}

    def add_monitor_repo(self, repo_name):
        self.repos_to_monitor.append(repo_name)

    def add_monitor_repo_list(self, lst):
        self.repos_to_monitor.extend(lst)

    def collect_baseline(self):
        for repo_name in self.repos_to_monitor:
            cnt = DB_CLIENT.count_of(repo_name)
            self.baseline[repo_name] = cnt

    def collect_info_after_update(self, after: Date):
        query_before = {'update_time': {'$lte': after.as_str_with_time()}}
        query_after = {'update_time': {'$gt': after.as_str_with_time()}}
        info_dict = self.info_after
        for repo_name in self.repos_to_monitor:
            cnt_before = DB_CLIENT.count_of(repo_name, query=query_before)
            cnt_after = DB_CLIENT.count_of(repo_name, query=query_after)
            info_dict[repo_name] = {
                'cnt_before': cnt_before,
                'cnt_after': cnt_after,
            }

    def report_changes(self):
        if len(self.baseline) < 1 or len(self.info_after) < 1:
            print("Invalid request, no baseline or info after change collected")
        for k, v in self.baseline.items():
            cnt_changed = self.info_after[k]
            print('## Changes of {} ##'.format(k))
            if cnt_changed is None or len(cnt_changed) < 1:
                print('No info after collected for {}'.format(k))
                continue
            if v != cnt_changed['cnt_before']:
                print('Warning: base cnt changed: {} -> {}'.format(v, cnt_changed['cnt_before']))
            print('updated cnt: {}'.format(cnt_changed['cnt_after']))


if __name__ == '__main__':
    dataService = SECURITY_DATA_READ_SERVICE
    verifier = SecurityDataUpdateMonitor()
    verifier.collect_baseline()
    verifier.collect_info_after_update(after=Date())
    verifier.report_changes()




from gcandle.core.sell_strategy.AbstractSellStrategy import (AbstractSellStrategy, SPR_KEY, SELL_STRATEGY_KEY)
import pandas as pd
import numpy as np


class CombinedSellStrategy(AbstractSellStrategy):

    def __init__(self, name):
        super().__init__(name)
        self.strategy_chain = []
        self.sell_history = {}

    def append(self, strategy):
        self.strategy_chain.append(strategy)
        self.sell_history[strategy.get_name()] = 0
        return self

    def add_sold(self, spr):
        counts = spr.groupby(SELL_STRATEGY_KEY).count()
        for row in counts.itertuples():
            self.sell_history[row.Index] += row.spr

    def get_history(self):
        return self.sell_history

    @staticmethod
    def get_empty_price_df(data):
        return pd.DataFrame(
            data={SPR_KEY: np.zeros(len(data)), SELL_STRATEGY_KEY: ["None"] * len(data)},
            index=data.index
        )

    # override
    def get_prices(self, data) -> pd.DataFrame:
        spr = CombinedSellStrategy.get_empty_price_df(data)
        for item in self.strategy_chain:
            if len(spr.loc[spr[SPR_KEY] == 0]) <= 0:
                break
            tmp_prices = item.get_prices(data)
            tmp_prices[SPR_KEY].fillna(0, inplace=True)
            if len(tmp_prices.loc[tmp_prices[SPR_KEY] > 0]) > 0:
                spr.loc[spr[SPR_KEY] == 0, [SPR_KEY, SELL_STRATEGY_KEY]] = tmp_prices[[SPR_KEY, SELL_STRATEGY_KEY]]

        if len(spr.loc[spr[SPR_KEY] > 0]) > 0:
            self.add_sold(spr.loc[spr[SPR_KEY] > 0])
        return spr



import pandas as pd

import gcandle.indicator.price_amount_features as fp_features
from gcandle.core.indicator_service.fp_mark_service import FPMarkService
from gcandle.core.indicator_service.slave_indicator_service import SlaveIndicatorService
from gcandle.data.se_types import SeFreq

FP_FEATURE_REPO_NAME_PREFIX = 'fp_features_'


class FpFeatureService(SlaveIndicatorService):
    def __init__(self, fpd):
        super().__init__()
        self.fpd = fpd
        self.fp_mark_service = FPMarkService(fpd, freq=SeFreq.DAY)

    #override
    def repo_name(self):
        return FP_FEATURE_REPO_NAME_PREFIX + str(self.fpd) + '_' + self.master_service.get_key()

    #override
    def days_to_update(self):
        return int(self.fpd * 15)

    #override
    def compute_indicator(self, data):
        codes = data.index.levels[1].unique().tolist()
        start = data.index.levels[0].min()
        end = data.index.levels[0].max()
        fpd = self.fpd
        fp_mark_data = self.fp_mark_service.read_by_dates(codes, start, end=end)
        if fp_mark_data is None or fp_mark_data.shape[0] < 1:
            return None

        data = pd.concat([data, fp_mark_data], axis=1)
        data = data.groupby(level=1).apply(fp_features.long_pivot_chg_features, fpd)
        return data


if __name__ == '__main__':
    print("This is fp features common db utils")


from gcandle.data.se_types import SeType
from gcandle.objects.data_service_objects import SECURITY_DATA_READ_SERVICE


class RiskIndicator:
    def __init__(self, account):
        self.bm_data = None
        self.bm_code = '000300'
        self.account = account
        self.assets = account.assets

    def _read_bm_data(self):
        self.bm_data = SECURITY_DATA_READ_SERVICE.read_security_data_for_single(
            self.bm_code,
            self.account.start_date,
            self.account.end_date,
            typ=SeType.Index
        )

    def _cacl_bm_assets(self):
        pass

    def benchmark_annualized_returns(self):
        pass

    def annualized_returns(self):
        pass

    def alpha(self):
        pass

    def beta(self):
        pass

    def sharp(self):
        pass



import pandas as pd
from gcandle.data.SecurityFetcher import SecurityFetcher
from gcandle.data.se_types import Exchange, SeFreq, SeType
from gcandle.utils.date_time_utils import Date


CSV_FILE_DIR = '/home/ych/stock/'

MAP_EX_TO_CSV_PREFIX = {
    Exchange.SZSE: 'SZ',
    Exchange.SHSE: 'SH'
}

class CsvSecurityFetcher(SecurityFetcher):
    def __init__(self):
        super().__init__()

    def fetch_security_list(self):
        print('Not implemented')
        pass

    def file_name(self, code, freq):
        exchange = Exchange.from_type_and_code(SeType.Stock, code)
        return CSV_FILE_DIR + 'stock_' + freq.value + '/' + MAP_EX_TO_CSV_PREFIX[exchange] + code + '.CSV'

    def fetch_bars(self, typ: SeType, code, start, end, freq):
        try:
            cols = ['date', 'time', 'open', 'high', 'low', 'close', 'vol', 'amount']
            data = pd.read_csv(self.file_name(code, freq), names=cols)
            if len(data) < 1:
                return None

            if freq.is_minutes():
                data['datetime'] = data['date'] + ' ' + data['time']
                data = data.assign(code=str(code)) \
                    .set_index('datetime', drop=True, inplace=False)
            return data
        except Exception as e:
            print('exception caught: {}'.format(e))




import pandas as pd
import numpy as np

SPR_KEY = 'spr'
SELL_STRATEGY_KEY = 'sell_strategy'

class AbstractSellStrategy:

    def __init__(self, name):
        self.name = name
        self.hold_range = [0, 10000]
        self.exclude_filters = []
        self.extra_filters = []

    def get_name(self):
        return self.name

    def get_price_and_name(self, prices):
        res = pd.DataFrame(data={SPR_KEY: prices, SELL_STRATEGY_KEY: [self.name] * len(prices)},
                            index=prices.index
                            )
        if len(res.loc[res[SPR_KEY] <= 0]) > 0:
            res.loc[res[SPR_KEY] <= 0, SELL_STRATEGY_KEY] = ""
        return res

    def set_hold_range(self, left, right=10000):
        self.hold_range = [left, right]
        self.add_extra_filter(
            lambda x : (left <= x.hold_days) & (x.hold_days <= right)
        )
        return self

    def clear_exclude_rules(self):
        self.exclude_filters.clear()
        return self

    def clear_extra_rules(self):
        self.extra_filters.clear()
        return self

    def add_exclude_filter(self, f):
        self.exclude_filters.append(f)
        return self

    def add_extra_filter(self, f):
        self.extra_filters.append(f)
        return self

    def get_prices(self, data) -> pd.DataFrame:
        prices = self.calc_prices(data)
        matched = pd.Series(data=[True] * len(prices), index=prices.index)
        for f in self.exclude_filters:
            excluded = f(data)
            matched &= ~excluded
        for f in self.extra_filters:
            matched &= f(data)

        if len(prices.loc[~matched]) > 0:
            prices.loc[~matched] = 0
        return self.get_price_and_name(prices)

    def calc_prices(self, data) -> pd.Series:
        #not implemented. should return series with same index of data
        pass



from gcandle.core.sell_strategy.AbstractSellStrategy import AbstractSellStrategy
import pandas as pd


class TargetPriceSellStrategy(AbstractSellStrategy):
    def __init__(self):
        super().__init__('TargetPrice')
        self.target_price_func = None
        self.add_exclude_filter(
            lambda x : x.open / x.pc >= 1.099
        )
        self.higher_than_pc = 0
        self.higher_than_bpr = 0

    def set_target_price_func(self, func):
        self.target_price_func = func
        return self

    def set_higher_than_pc(self, rate):
        self.higher_than_pc = rate
        return self

    def set_higher_than_bpr(self, rate):
        self.higher_than_bpr = rate
        return self

    #override
    def calc_prices(self, data):
        my_prices = pd.Series([0] * len(data), index=data.index)
        if self.target_price_func is not None:
            my_prices = data.apply(self.target_price_func, axis=1)
        elif self.higher_than_bpr > 0 and self.higher_than_pc > 0:
            tmp_df = pd.concat([data.pc * self.higher_than_pc, data.bpr * self.higher_than_bpr], axis=1)
            my_prices = tmp_df.min(axis=1)
        elif self.higher_than_bpr > 0:
            my_prices = data.bpr * self.higher_than_bpr
        elif self.higher_than_pc > 0:
            my_prices = data.pc * self.higher_than_pc
        else:
            print("Invalid target prices strategy, no function or higher rate is set")

        target_too_high = (data.high < my_prices)
        my_prices.loc[target_too_high] = 0
        my_prices.loc[(my_prices > 0) & (my_prices < data.open)] = data.open
        return my_prices




import json
import os


MONGO_URI = 'mongodb://localhost:27017'
DATABASE_NAME = 'gcandle'


class GcandleConfig:
    def __init__(self):
        self.mongodb_uri = MONGO_URI
        self.mongodb_name = DATABASE_NAME
        self.load_config()

    def set_mongodb_uri(self, uri):
        self.mongodb_uri = uri

    def set_mongodb_name(self, name):
        self.mongodb_name = name

    def load_config(self):
        CONFIG_FILE = os.getenv('GCANDLE_CONFIG_FILE')
        if CONFIG_FILE is None:
            HOME_DIR = os.getenv("HOME")
            if HOME_DIR is None:
                print('Cannot get $HOME dir, default config will be used.')
                return self
            CONFIG_FILE = HOME_DIR + '/' + '.gcandle.json'
        try:
            with open(CONFIG_FILE, 'r') as config_file:
                data = json.load(config_file)
                print("Config data: {}".format(data))
                if "mongodb_uri" in data:
                    self.mongodb_uri = data['mongodb_uri']
                if "mongodb_name" in data:
                    self.mongodb_name = data['mongodb_name']
        except Exception as e:
            print('Failed to open config file {} with error: {}'
                  .format(CONFIG_FILE, e))
            print(' default config will be used')


GCANDLE_CONFIG = GcandleConfig()


if __name__ == '__main__':
    print(GCANDLE_CONFIG.mongodb_uri, GCANDLE_CONFIG.mongodb_name)



from gcandle.data.se_types import SeType, SeFreq
from gcandle.objects.data_service_objects import TDX_SECURITY_DATA_UPDATE_SERVICE
from gcandle.data.SecurityDataUpdateMonitor import SecurityDataUpdateMonitor
from gcandle.utils.date_time_utils import Date
import time


def fetch_and_update_by_tdx():
    dataService = TDX_SECURITY_DATA_UPDATE_SERVICE
    monitor = SecurityDataUpdateMonitor()
    a = Date()
    monitor.add_monitor_repo("security_list")
    monitor.add_monitor_repo("index_day")
    monitor.add_monitor_repo("stock_day")
    monitor.add_monitor_repo("stock_5min")
    monitor.add_monitor_repo("stock_30min")
    monitor.add_monitor_repo("security_xdxr")
    monitor.collect_baseline()
    dataService.update_code_list()
    dataService.update_bars(typ=SeType.Index, freq=SeFreq.DAY)
    dataService.update_bars(typ=SeType.Stock, freq=SeFreq.DAY)
    codes = dataService.read_security_codes(typ=SeType.Stock)
    dataService.refetch_and_save_xdxr(codes)
    monitor.collect_info_after_update(a)
    b = Date()
    print("time used: {} seconds".format(b.delta_to(a).seconds))
    monitor.report_changes()


if __name__ == '__main__':
    fetch_and_update_by_tdx()


from gcandle.core.backtest.seposition import SePosition
from collections import namedtuple

PositionKey = namedtuple('Position', ['date', 'code'])


class AccountPositionStock:
    def __init__(self):
        self.history_positions = {}
        self.active_positions = {}
        self.daily_pos_snapshot = []

    def get_active_position_days(self, code):
        assert code in self.active_positions, "Cannot get days for non-existing code {}".format(code)
        return self.active_positions[code].get_days()

    def open_position(self, price, amount, data):
        code = data.code
        if code in self.active_positions:
            self.active_positions[code].open(price, amount, data)
        else:
            self.active_positions[code] = SePosition(code).open(price, amount, data)

    def close_position(self, price, data):
        date = data.date
        code = data.code
        assert code in self.active_positions, "invalid parameter, {} not in active positions".format(code)
        position = self.active_positions[code]
        position.close_all(price, data)
        key = PositionKey(date=data.date, code=data.code)
        self.history_positions[key] = position
        del self.active_positions[code]

    def close_trade(self, data):
        # 收盘: 增加每日持仓的记录, 包含 date, code, close, amount, se_value, days
        date = data.date[0]
        for code, pos in self.active_positions.items():
            item = data[data.code == code]
            pos.update_status(date, item)
            self.daily_pos_snapshot.append(pos.get_snapshot())

    def get_daily_positions(self):
        return self.daily_pos_snapshot

    def get_all_positions(self):
        return self.active_positions, self.history_positions

    def get_won_positions(self):
        return [p for p in self.history_positions.values() if p.profit() > 0]

    def get_lost_positions(self):
        return [p for p in self.history_positions.values() if p.profit() < 0]

    def win_rate(self):
        return round(len(self.get_won_positions()) / len(self.history_positions), 3)



import math
import numpy as np
import pandas as pd

def increase_limit(item):
    return item.close / item.pc >= 1.099


def decrease_limit(item):
    return item.close / item.pc <= 0.9001


def ultimate_increase(item):
    return item.close > item.pc and (item.open == item.close == item.high == item.low)


def ultimate_decrease(item):
    return item.close < item.pc and (item.open == item.close == item.high == item.low)



def EMA(Series, N):
    return pd.Series.ewm(Series, span=N, min_periods=N - 1, adjust=True).mean()


def MA(Series, N):
    return pd.Series.rolling(Series, N).mean()


def SMA(Series, N, M=1):
    ret = []
    i = 1
    length = len(Series)
    # 跳过X中前面几个 nan 值
    while i < length:
        if np.isnan(Series.iloc[i]):
            i += 1
        else:
            break
    preY = Series.iloc[i]  # Y'
    ret.append(preY)
    while i < length:
        Y = (M * Series.iloc[i] + (N - M) * preY) / float(N)
        ret.append(Y)
        preY = Y
        i += 1
    return pd.Series(ret, index=Series.tail(len(ret)).index)


def DIFF(Series, N=1):
    return pd.Series(Series).diff(N)


def HHV(Series, N):
    return pd.Series(Series).rolling(N).max()


def LLV(Series, N):
    return pd.Series(Series).rolling(N).min()


def SUM(Series, N):
    return pd.Series.rolling(Series, N).sum()


def ABS(Series):
    return abs(Series)


def MAX(A, B):
    var = IF(A > B, A, B)
    return var


def MIN(A, B):
    var = IF(A < B, A, B)
    return var


def SINGLE_CROSS(A, B):
    if A.iloc[-2] < B.iloc[-2] and A.iloc[-1] > B.iloc[-1]:
        return True
    else:
        return False


def CROSS(A, B):
    var = np.where(A < B, 1, 0)
    return (pd.Series(var, index=A.index).diff() < 0).apply(int)


def COUNT(COND, N):
    return pd.Series(np.where(COND, 1, 0), index=COND.index).rolling(N).sum()


def IF(COND, V1, V2):
    var = np.where(COND, V1, V2)
    return pd.Series(var, index=V1.index)


def IFAND(COND1, COND2, V1, V2):
    var = np.where(np.logical_and(COND1, COND2), V1, V2)
    return pd.Series(var, index=V1.index)


def IFOR(COND1, COND2, V1, V2):
    var = np.where(np.logical_or(COND1, COND2), V1, V2)
    return pd.Series(var, index=V1.index)


def REF(Series, N):
    var = Series.diff(N)
    var = Series - var
    return var


def LAST(COND, N1, N2):
    N2 = 1 if N2 == 0 else N2
    assert N2 > 0
    assert N1 > N2
    return COND.iloc[-N1:-N2].all()


def STD(Series, N):
    return pd.Series.rolling(Series, N).std()


def AVEDEV(Series, N):
    return Series.rolling(N).apply(lambda x: (np.abs(x - x.mean())).mean(), raw=True)


def MACD(CLOSE, FAST=12, SLOW=26, MID=9):
    EMAFAST = EMA(CLOSE, FAST)
    EMASLOW = EMA(CLOSE, SLOW)
    DIFF = EMAFAST - EMASLOW
    DEA = EMA(DIFF, MID)
    MACD = (DIFF - DEA) * 2
    DICT = {'DIFF': DIFF, 'DEA': DEA, 'MACD': MACD}
    VAR = pd.DataFrame(DICT)
    return VAR


def BBIBOLL(Series, N1, N2, N3, N4, N, M):

    bbiboll = BBI(Series, N1, N2, N3, N4)
    UPER = bbiboll + M * STD(bbiboll, N)
    DOWN = bbiboll - M * STD(bbiboll, N)
    DICT = {'BBIBOLL': bbiboll, 'UPER': UPER, 'DOWN': DOWN}
    VAR = pd.DataFrame(DICT)
    return VAR


def BBI(Series, N1, N2, N3, N4):
    bbi = (MA(Series, N1) + MA(Series, N2) +
           MA(Series, N3) + MA(Series, N4)) / 4
    DICT = {'BBI': bbi}
    VAR = pd.DataFrame(DICT)
    return VAR


def BARLAST(cond, yes=True):
    if isinstance(cond.index, pd.MultiIndex):
        return len(cond) - cond.index.levels[0].tolist().index(cond[cond != yes].index[-1][0]) - 1
    elif isinstance(cond.index, pd.DatetimeIndex):
        return len(cond) - cond.index.tolist().index(cond[cond != yes].index[-1]) - 1


XARROUND = lambda x, y: np.round(y * (round(x / y - math.floor(x / y) + 0.00000000001) + math.floor(x / y)), 2)




from gcandle.data.se_types import SeType, SeFreq
from gcandle.objects.data_service_objects import TDX_SECURITY_DATA_UPDATE_SERVICE
from gcandle.data.SecurityDataUpdateMonitor import SecurityDataUpdateMonitor
from gcandle.utils.date_time_utils import Date
import time


def fetch_and_update_min_bars_by_tdx():
    dataService = TDX_SECURITY_DATA_UPDATE_SERVICE
    monitor = SecurityDataUpdateMonitor()
    codes = TDX_SECURITY_DATA_UPDATE_SERVICE.read_security_codes()
    a = Date()
    # monitor.add_monitor_repo("stock_1min")
    # monitor.add_monitor_repo("stock_5min")
    # monitor.add_monitor_repo("stock_30min")
    # monitor.collect_baseline()
    # codes = ['000001', '300343'] + codes[:100]
    dataService.update_bars(typ=SeType.Stock, freq=SeFreq.ONE_MIN, codes=codes)
    dataService.update_bars(typ=SeType.Stock, freq=SeFreq.FIVE_MIN, codes=codes)
    dataService.update_bars(typ=SeType.Stock, freq=SeFreq.THIRTY_MIN, codes=None)
    # monitor.collect_info_after_update(a)
    b = Date()
    print("time used: {} seconds".format(b.delta_to(a).seconds))
    # monitor.report_changes()


if __name__ == '__main__':
    fetch_and_update_min_bars_by_tdx()


from enum import Enum


class PositionDir(Enum):
    OPEN = 1,
    CLOSE = 2


class SeTrade:
    def __init__(self, dir: PositionDir, date: str, price: float, amount: int, low: float, high: float):
        self.dir = dir
        self.date = date
        self.price = price
        self.amount = amount
        self.low = low
        self.high = high


class SePosition:

    def __init__(self, code):
        self.open_date = None
        self.clear_date = None
        self.today = None
        self.code = code
        self.amount = 0
        self.se_value = 0
        self.days = 0
        self.cleared = False
        self.trade_history = [] # SeTrade

    def get_amount(self):
        return self.amount

    def update_status(self, date, item):
        self.increase_day()
        self.today = date
        if item is not None and len(item) > 0:
            item = item.squeeze()
            self.se_value = self.amount * item.close

    def get_snapshot(self):
        return [self.today, self.code, self.amount, self.se_value, self.days]

    def profit(self):
        if len(self.trade_history) < 1:
            return 0
        total_pay = 0
        total_got = 0
        for r in self.trade_history:
            if r.dir == PositionDir.OPEN:
                total_pay += r.price * r.amount
            else:
                total_got += r.price * r.amount
        return (total_got - total_pay) / total_pay

    def to_dict(self):
        return {
            open: self.open_date
        }

    def get_trade_history(self):
        return self.trade_history

    def get_days(self):
        return self.days

    def increase_day(self, days=1):
        self.days += days
        return self

    def open(self, price, amount, data):
        assert amount > 0 and price > 0, "Invalide parameter for creating position. price={} amount={}".format(price, amount)
        assert self.cleared == False, "Cannot open for cleared"
        assert data.code == self.code, "Invalid parameter, code don't match {} {}".format(self.code, data.code)
        self.trade_history.append(SeTrade(
            PositionDir.OPEN,
            data.date,
            price,
            amount,
            data.low,
            data.high
        ))
        self.amount += amount
        if self.open_date is None:
            self.open_date = data.date
        return self

    def close(self, price, amount, data):
        if data.low > price or price > data.high:
            print("Warning: close price out of bound of data {}-{} ".format(data.low, data.high))
        assert amount > 0, "Invalid parameter of amount {}".format(amount)
        assert self.amount >= amount, "Position is less than amount to close"
        self.trade_history.append(SeTrade(
            PositionDir.CLOSE,
            data.date,
            price,
            self.amount,
            data.low,
            data.high
        ))
        self.amount -= amount
        if self.amount == 0:
            self.cleared = True
            self.clear_date = data.date
        return self

    def close_all(self, price, data):
        return self.close(price, self.amount, data)




from gcandle.objects.Indicator_service_objects import DAY_FP_MARK_SERVICES
from gcandle.objects.data_service_objects import SECURITY_DATA_READ_SERVICE


def update_all_foot_peak_marks():
    codes = SECURITY_DATA_READ_SERVICE.read_security_codes()
    for fp_mark_service in DAY_FP_MARK_SERVICES:
        fp_mark_service.update_all_codes(codes)


def recreate_all_foot_peak_marks():
    codes = SECURITY_DATA_READ_SERVICE.read_security_codes()
    for fp_mark_service in DAY_FP_MARK_SERVICES:
        fp_mark_service.recreate_all_codes(codes)


if __name__=="__main__":
    recreate_all_foot_peak_marks()
    # update_all_foot_peak_marks()
    # update_foot_peak_marks('300811', 20, '2019-01-01', '2020-07-30')
    pass



from gcandle.core.indicator_service.indicator_db_service import IndicatorDBService
from gcandle.objects.basic_objects import DB_CLIENT
from gcandle.objects.data_service_objects import SECURITY_DATA_READ_SERVICE
from gcandle.utils.date_time_utils import Date
from gcandle.data.se_types import SeFreq, SeType
import gcandle.utils.multi_process_pool as executor


INDEPENDENT_INDICATOR_REPO_PREFIX = 'indicator_'
DEFAULT_MIN_START_DATE = '2006-01-01'
DEFAULT_DAYS_TO_UPDATE = 15


class IndicatorService:
    def __init__(self, name: str, indicator_func, freq: SeFreq, days_to_update: int = DEFAULT_DAYS_TO_UPDATE):
        self.name = name
        self.indicator_func = indicator_func
        self.freq = freq
        self.days_to_update = days_to_update
        self.db_service = IndicatorDBService(name, freq)

    #to be overridden by subclass
    def pre_save(self, data):
        return data

    def update_for_single_code(self, code, start, end):
        data = self.read_data_and_calc_indicators(code, start, end)
        self.db_service.save(data)

    def do_update_by_dates_for_code(self, code, start, end):
        data = self.read_data_and_calc_indicators(code, start, end)
        self.db_service.replace_by_dates(data, code, start, end)

    def read_data_and_calc_indicators(self, code, start, end):
        if self.freq.is_minutes():
            data = SECURITY_DATA_READ_SERVICE.read_security_data_for_single(code, start, end, freq=self.freq, typ=SeType.Stock)
        else:
            data = SECURITY_DATA_READ_SERVICE.read_qfq_security_data_for_single(code, start, end)
        if data is None or data.shape[0] < 10:
            return

        data = self.indicator_func(data)
        if data is None or len(data) < 1:
            return None
        else:
            return self.pre_save(data)

    def update_all_codes(self, codes):
        end = Date()
        start = end.get_before(self.days_to_update)
        executor.execute_tasks(
            codes,
            self.update_for_single_code,
            start.as_str(),
            end.as_str()
        )

    def update_by_dates(self, codes, start, end):
        executor.execute_tasks(
            codes,
            self.do_update_by_dates_for_code,
            start,
            end
        )

    def recreate_all_codes(self, codes, create_start_date=DEFAULT_MIN_START_DATE):
        if codes is None:
            if self.db_service.confirm_and_drop():
                codes = SECURITY_DATA_READ_SERVICE.read_security_codes()
            else:
                return
        else:
            self.db_service.remove_all(codes)

        end = Date()
        start = Date.from_str(create_start_date)
        executor.execute_tasks(
            codes,
            self.update_for_single_code,
            start.as_str(),
            end.as_str()
        )

    def read_by_dates(
            self,
            codes,
            start='all',
            end=None
    ):
        return self.db_service.read_by_dates(codes, start, end)

    def read_by_data(self, data):
        return self.db_service.read_by_data(data)

    def read_all_codes(self):
        return self.db_service.read_all_codes()

    def __confirm_drop(self):
        confirm = input("确认删除所有的数据？请确认[yes/no]:")
        return confirm == 'yes'




from gcandle.core.sell_strategy.AbstractSellStrategy import AbstractSellStrategy


class SellAtOpenStrategy(AbstractSellStrategy):
    def __init__(self):
        super().__init__('SellAtOpen')
        self.add_exclude_filter(
            lambda data : data.open / data.pc >= 1.099
        )

    #override
    def calc_prices(self, data):
        my_prices = data.open.copy()
        return my_prices




from gcandle.data.csv.CsvSecurityFetcher import CsvSecurityFetcher
from gcandle.data.tdx.TdxSecurityFetcher import TdxSecurityFetcher
from gcandle.data.SecurityDataService import SecurityDataService

SECURITY_DATA_READ_SERVICE = SecurityDataService()

TDX_SECURITY_DATA_UPDATE_SERVICE = SecurityDataService(TdxSecurityFetcher())

CSV_SECURITY_DATA_SERVICE = SecurityDataService(CsvSecurityFetcher())

if __name__ == '__main__':
    print('Defines objects that to be used for security data service')



from concurrent.futures import Executor, ProcessPoolExecutor, ThreadPoolExecutor, as_completed, Future
from gcandle.objects.basic_objects import DB_CLIENT

import time


def pool_init():
    DB_CLIENT.init()


class CodeListTaskExecutor:
    def __init__(self, executor_cls, n_workers=8):
        self.n_workers = n_workers
        self.executor = executor_cls(max_workers=n_workers)

    def execute(self, codes: list, task_func, *args):
        executor: Executor = self.executor
        start = time.time()
        futures_of_task = {
            self.executor.submit(task_func, c, *args) for c in codes
        }
        print('submitted')
        i = 0
        ntotal = len(codes)
        for future in as_completed(futures_of_task):
            i = i+1
            if i % 100 == 0:
                print("Task execution progress: {:.2f}% of {}".format(i * 100.0 / ntotal, ntotal))

        print("Task all done")
        executor.shutdown()




import pandas as pd
import gcandle.utils.date_time_utils as date_time_utils
from pytdx.hq import TdxHq_API
from gcandle.data.SecurityFetcher import SecurityFetcher
from gcandle.data.se_types import Exchange, SeFreq, SeType

#map to a list with [frequence number, count factor of days]
map_freq_to_tdx_number = {
    SeFreq.ONE_MIN: [8, 240],
    SeFreq.FIVE_MIN: [0, 48],
    SeFreq.FIFTEEN_MIN: [1, 16],
    SeFreq.THIRTY_MIN: [2, 8],
    SeFreq.DAY: [9, 1],
    SeFreq.WEEK: [5, 0.2],
    SeFreq.MONTH: [6, 0.02],
    SeFreq.YEAR: [11, 0.005]
}

map_exchange_to_tdx_number = {
    Exchange.SZSE: 0,
    Exchange.SHSE: 1
}

MAX_DAYS = 4000
BATCH_SIZE = 800
STOCK_DATA_SERVER = {
    "ip": "221.194.181.176",
    "port": 7709
}

class TdxSecurityFetcher(SecurityFetcher):
    def __init__(self):
        super().__init__()
        self.server = STOCK_DATA_SERVER

    def __do_fetch_security_list(self, api: TdxHq_API, exchange: Exchange):
        arr = []
        start = 0
        ex_nbr = map_exchange_to_tdx_number[exchange]
        count = api.get_security_count(ex_nbr)
        while start < count:
            tmp = api.get_security_list(ex_nbr, start)
            if tmp is None:
                break
            sz = len(tmp)
            if sz < 1:
                break
            start += sz
            arr.extend(tmp)

        print('total fetched rows: {}'.format(start))
        df = api.to_df(arr).assign(ex=exchange.name)
        df = df.assign(
            stype=df.code.apply(exchange.code_to_type_str)
        ).set_index(['code', 'ex'], drop=True)
        return df[df.stype.notnull()]

    def fetch_security_list(self):
        api = TdxHq_API()
        with api.connect(self.server['ip'], self.server['port'], time_out=3):
            data = pd.concat([
                self.__do_fetch_security_list(api, ex) for ex in Exchange], axis=0
            )
        return data

    def fetch_bars(self, typ: SeType, code, start, end, freq):
        try:
            tdxApi = TdxHq_API()
            ex_nbr = map_exchange_to_tdx_number[Exchange.from_type_and_code(typ, code)]
            if typ == SeType.Stock or typ == SeType.ETF:
                tdx_fetch_func = tdxApi.get_security_bars
            elif typ == SeType.Index:
                tdx_fetch_func = tdxApi.get_index_bars
            else:
                raise Exception('Unsupported se type')

            #timeout is in seconds.
            with tdxApi.connect(self.server['ip'], self.server['port'], time_out=3):
                frequence = map_freq_to_tdx_number[freq][0]
                days_between = date_time_utils.Date.from_str(end).delta_to(start).days + 1
                if days_between > MAX_DAYS:
                    days_between = MAX_DAYS
                max_count = int(map_freq_to_tdx_number[freq][1] * days_between)
                tmp_list = []
                n_start = 0
                while n_start < max_count:
                    fetch_size = min(max_count - n_start, BATCH_SIZE)
                    tmp_res = tdx_fetch_func(
                        frequence,
                        ex_nbr,
                        code, n_start, fetch_size
                    )
                    if tmp_res is None or len(tmp_res) < 1:
                        break
                    n_start += len(tmp_res)
                    tmp_list.extend(tmp_res)
                if n_start == 0:
                    print('No rows fetched for {}'.format(code))
                # print('total fetched rows for {}: {}'.format(code, n_start))
                data = tdxApi.to_df(tmp_list)

                if len(data) < 1:
                    return None
                data = data[data['open'] != 0]
                cols_to_drop = ['year', 'month', 'day']

                if freq.is_minutes():
                    data = data.assign(datetime=data['datetime'].apply(lambda x: str(x[0:19])),
                                       date=data['datetime'].apply(lambda x: str(x[0:10])),
                                       code=str(code)) \
                        .set_index('datetime', drop=True, inplace=False) \
                        .drop(cols_to_drop, axis=1)
                else:
                    cols_to_drop.extend(['datetime', 'hour', 'minute'])
                    data = data.assign(datetime=data['datetime'],
                                       date=data['datetime'].apply(lambda x: str(x[0:10])),
                                       code=str(code)) \
                        .set_index('date', drop=True, inplace=False) \
                        .drop(cols_to_drop, axis=1)
                return data
        except Exception as e:
            print('exception caught: {}'.format(e))

    def fetch_xdxr(self, code):
        ip, port = self.server['ip'], self.server['port']
        api = TdxHq_API()
        ex_nbr = map_exchange_to_tdx_number[Exchange.from_type_and_code(SeType.Stock, code)]
        with api.connect(ip, port, time_out=3):
            data = api.to_df(api.get_xdxr_info(ex_nbr, code))
            if len(data) >= 1:
                data = data \
                    .assign(date=pd.to_datetime(data[['year', 'month', 'day']])) \
                    .drop(['year', 'month', 'day'], axis=1) \
                    .assign(code=str(code))
                return data.assign(date=data['date'].apply(lambda x: str(x)[0:10]))\
                    .set_index('date', drop=True, inplace=False)
            else:
                return None



from gcandle.objects.basic_objects import DB_CLIENT
from gcandle.objects.data_service_objects import SECURITY_DATA_READ_SERVICE
from gcandle.utils.date_time_utils import Date
from gcandle.data.se_types import SeFreq, SeType
import gcandle.utils.multi_process_pool as executor


INDEPENDENT_INDICATOR_REPO_PREFIX = 'indicator_'
DEFAULT_MIN_START_DATE = '2006-01-01'
DEFAULT_DAYS_TO_UPDATE = 15


class IndicatorDBService:
    def __init__(self, name: str, freq: SeFreq):
        self.name = name
        self.freq = freq

    def replace_by_dates(self, data, code, start, end):
        DB_CLIENT.delete_by_code_and_dates(self.repo_name(), code, start, end)
        if data is not None and len(data) > 0:
            DB_CLIENT.create_index(self.repo_name(), self.freq.keys_to_index())
            repo_name = self.repo_name()
            DB_CLIENT.replace_data_by_date_range(data, repo_name, start, end)

    def save(self, data):
        DB_CLIENT.create_index(self.repo_name(), self.freq.keys_to_index())
        if data is not None and len(data) > 0:
            repo_name = self.repo_name()
            DB_CLIENT.update_data_append_newer(data, repo_name)

    def repo_name(self):
        return INDEPENDENT_INDICATOR_REPO_PREFIX + self.name

    def read_by_dates(
            self,
            codes,
            start='all',
            end=None
    ):
        repo_name = self.repo_name()
        return DB_CLIENT.read_data_by_codes_and_date(repo_name, codes, start, end=end)

    def read_by_data(self, data):
        codes = data.index.levels[1].unique().tolist()
        start = data.date.min()
        end = data.date.max()
        return self.read_by_dates(codes, start, end)

    def read_all_codes(self):
        repo_name = self.repo_name()
        return DB_CLIENT.read_codes_from_db(repo_name)

    def __confirm_drop(self):
        confirm = input("确认删除所有的数据？请确认[yes/no]:")
        return confirm == 'yes'

    def confirm_and_drop(self):
        if self.__confirm_drop():
            DB_CLIENT.drop(self.repo_name())
            print('{} dropped'.format(self.repo_name()))
            return True
        else:
            print("操作已取消。")
            return False

    def remove_all(self, codes):
        DB_CLIENT.remove_all_from(self.repo_name(), query={'code': {'$in': codes}})



from enum import Enum

import numpy as np


class FOOT_PEAK_TYPE(Enum):
    FOOT = 1
    PEAK = 2

FP_CHANGES_COUNT=3


def price_foot_peak_mark(data, fpd):
    data['day_idx'] = np.arange(1, data.shape[0] + 1)

    data['highest_b_' + str(fpd)] = data.high.rolling(fpd+1, min_periods=1).max()
    data['lowest_b_' + str(fpd)] = data.low.rolling(fpd+1, min_periods=1).min()
    data['highest_a_' + str(fpd)] = data['highest_b_' + str(fpd)].shift(-fpd)
    data['lowest_a_' + str(fpd)] = data['lowest_b_' + str(fpd)].shift(-fpd)

    data['max_around_' + str(fpd)] = data[['highest_b_' + str(fpd), 'highest_a_' + str(fpd)]].max(axis=1, skipna=False)
    data['min_around_' + str(fpd)] = data[['lowest_b_' + str(fpd), 'lowest_a_' + str(fpd)]].min(axis=1, skipna=False)

    foot_filter = data['min_around_' + str(fpd)] == data['low']
    peak_filter = data['max_around_' + str(fpd)] == data['high']
    data.loc[foot_filter, 'fp_type'] = FOOT_PEAK_TYPE.FOOT.value
    data.loc[peak_filter, 'fp_type'] = FOOT_PEAK_TYPE.PEAK.value
    fp_col = 'fp_around_' + str(fpd)
    data.loc[foot_filter | peak_filter, fp_col] = int(fpd)

    # drop duplicates
    if data.loc[data[fp_col] == fpd].shape[0] > 0:
        data.loc[data[fp_col] == int(fpd), 'last_fp'] = data['fp_type']
        data['last_fp'].fillna(method='ffill', inplace=True)
        data.loc[data[fp_col] == int(fpd), 'last_fp_idx'] = data['day_idx']
        data['last_fp_idx'].fillna(method='ffill', inplace=True)
        duplicates = (data[fp_col] == int(fpd)) & (data['last_fp'].shift(1) == data['fp_type']) & (
                data['day_idx'] - data['last_fp_idx'].shift(1) < fpd)
        # print(data.loc[duplicates, ['high', 'low', 'fp_type', fp_col]])
        data.loc[duplicates, fp_col] = np.nan
        data.drop(['last_fp', 'last_fp_idx'], axis=1, inplace=True)

    mask = (data['fp_around_' + str(fpd)].isna())
    data.loc[mask, 'fp_type'] = np.nan
    to_drop = [s for s in data.columns if 'max_' in s or 'min_' in s]
    data.drop(to_drop, axis=1, inplace=True)


def long_pivot_chg_features(data, fpd):

    data['day_idx'] = np.arange(1, data.shape[0] + 1)
    fp_around_col = 'fp_around_' + str(fpd)
    fp_rows = data[fp_around_col] == fpd

    chg_type = 'p' + str(fpd) + '_'
    n_points = FP_CHANGES_COUNT
    data.loc[fp_rows, 'fp_idx'] = data.loc[fp_rows, 'day_idx']
    data[chg_type + 'last_0_fp_price'] = data.close
    data[chg_type + 'last_0_fp_idx'] = data.day_idx

    #mark the pivot points
    cols = ['low', 'high', 'fp_idx', 'fp_type']
    for n in range(0, n_points):
        prefix = chg_type + 'last_' + str(n + 1) + '_'
        for col in cols:
            last_col = prefix + col
            data[last_col] = np.nan
            if n == 0:
                if data.loc[data[fp_around_col].shift(fpd) == fpd].shape[0] > 0:
                    data.loc[data[fp_around_col].shift(fpd) == fpd, last_col] = data[col].shift(fpd)
                # on fp day, last_col == col. So for fpd later, shift(1) should be the correct last_col.
                # data[last_col] = data[last_col].fillna(method='ffill')
            else:
                pre_last_col = chg_type + 'last_' + str(n) + '_' + col
                last_rows = data[pre_last_col].notna()
                data.loc[last_rows, last_col] = data.loc[last_rows, pre_last_col].shift(1)

    #fill:
    for n in range(1, n_points+1):
        prefix = chg_type + 'last_' + str(n) + '_'
        cols_to_adjust = [prefix + col for col in cols]
        data[cols_to_adjust] = data[cols_to_adjust].fillna(method='ffill')

        data[prefix + 'fp_price'] = np.nan
        if data.loc[data[prefix + 'fp_type'] == 1].shape[0] > 0:
            data.loc[data[prefix + 'fp_type'] == 1, prefix + 'fp_price'] = data[prefix + 'low']
        if data.loc[data[prefix  + 'fp_type'] == 2].shape[0] > 0:
            data.loc[data[prefix + 'fp_type'] == 2, prefix + 'fp_price'] = data[prefix + 'high']

        data[prefix + 'fp_dist'] = data.day_idx - data[prefix + 'fp_idx']
        data[prefix + 'fp_chg_len'] = data[chg_type + 'last_' + str(n-1) + '_fp_idx'] - data[prefix + 'fp_idx']
        data[prefix + 'fp_chg' ] = data[chg_type + 'last_' + str(n-1) + '_fp_price'] / data[prefix + 'fp_price']

    #special filling for ci xin
    treat_ci_xin_as_big_rise_fp(chg_type, data)

    to_drop = [s for s in data.columns if
               '_idx' in s or '_amount' in s or '_low' in s \
               or '_high' in s or 'shangshi' in s \
               or 'fp_around' in s or 'fp_type' in s \
               or 'fp_price' in s or 'last_0' in s\
               or 'close' in s]
    data.drop(to_drop, axis=1, inplace=True)
    return data


def treat_ci_xin_as_big_rise_fp(chg_type, data):
    data.loc[data.day_idx == 1, 'shangshi'] = data.open / 2
    data['shangshi'].fillna(method='ffill', inplace=True)
    ci_xin_has_only_one_fp_chg = (data[chg_type + 'last_1_fp_chg'].notna()) & (data[chg_type + 'last_2_fp_chg'].isna()) &\
                                 (data[chg_type + 'last_1_fp_idx'] != 1)

    if len(data.loc[ci_xin_has_only_one_fp_chg]) > 0:
        data.loc[ci_xin_has_only_one_fp_chg, chg_type + 'last_2_fp_chg'] = \
            data[chg_type + 'last_1_fp_price'] / data.shangshi
        data.loc[ci_xin_has_only_one_fp_chg, chg_type + 'last_2_fp_dist'] = \
            data.day_idx
        data.loc[ci_xin_has_only_one_fp_chg, chg_type + 'last_2_fp_chg_len'] = \
            data[chg_type + 'last_1_fp_idx']


def label_increase(data, days=8, threshold=1.3):
    col = 'a' + str(days) + '_max_prc'
    data[col] = data.close.shift(-days).rolling(days).max()
    data['win' + str(days)] = data[col] / data.close > threshold



from gcandle.core.sell_strategy.AbstractSellStrategy import AbstractSellStrategy
from gcandle.indicator.basic_indicators import (increase_limit, decrease_limit)


class SellAtCloseStrategy(AbstractSellStrategy):
    def __init__(self, name='SellAtClose'):
        super().__init__(name)
        self.add_exclude_filter(
            increase_limit
        ).add_exclude_filter(
            decrease_limit
        )

    #override
    def calc_prices(self, data):
        my_prices = data.close.copy()
        return my_prices




# coding:utf-8

from decimal import *
import numpy as np
import pandas as pd
from gcandle.objects.data_service_objects import SECURITY_DATA_READ_SERVICE

category = {
    '1': '除权除息', '2': '送配股上市', '3': '非流通股上市', '4': '未知股本变动', '5': '股本变化',
    '6': '增发新股', '7': '股份回购', '8': '增发新股上市', '9': '转配股上市', '10': '可转债上市',
    '11': '扩缩股', '12': '非流通股缩股', '13': '送认购权证', '14': '送认沽权证'}


def stock_to_qfq(data, verbose=False):
    start = str(data.index.levels[0].min())[0:10]
    end = str(data.index.levels[0].max())[0:10]

    code = data.index.levels[1][0]
    xdxr_data = SECURITY_DATA_READ_SERVICE.read_fq_info(code, start, end)

    return do_prices_qfq(
        bfq_data=data,
        xdxr_data=xdxr_data,
        verbose=verbose
    )


def do_prices_qfq(bfq_data, xdxr_data, verbose=False):
    if xdxr_data is None:
        print("{} has no xdxr data ".format(bfq_data.index.levels[1][0]))
        return bfq_data

    if len(xdxr_data) <= 0:
        print('Warn: no xdxr info, market value will be missing for {}'.format(bfq_data.index.levels[1][bfq_data.index.codes[1]]))
        return bfq_data

    data = bfq_data.assign(if_trade=True)

    # 复权股本
    # data = to_qfq_shares(data, xdxr_data)
    data = data.query('if_trade==True')

    #复权股价
    info = xdxr_data.query('category==1').loc[bfq_data.index[0]:bfq_data.index[-1]]
    if len(info) > 0:
        fq_columns = ['h', 'r']
        cummulate_rights(info)
        data = pd.concat(
            [
                data,
                info.loc[bfq_data.index[0]:bfq_data.index[-1], fq_columns]
            ],
            axis=1
        )
        rows_to_watch = None
        if verbose:
            rows_to_watch = data[fq_columns].notna().any(axis=1)
            rows_to_watch = rows_to_watch.replace(False, np.nan).\
                fillna(method='bfill', limit=3).fillna(method='ffill', limit=3).fillna(False).replace(1, True)
        data[fq_columns] = data[fq_columns].shift(-1).fillna(method='bfill').astype(float)
        data['h'].fillna(0, inplace=True)
        data['r'].fillna(1, inplace=True)

        # 如果中间有除权信息但是没有交易， if_trade 就会是False
        data['if_trade'].fillna(value=False, inplace=True)

        # 如果中间有除权信息但是没有交易，股价会用上个交易日的信息填充
        data = data.fillna(method='ffill')

        data = data.fillna(0)
        if verbose:
            print('before qfq, data:\n', data.loc[rows_to_watch, ['close', 'volume', 'h', 'r']])

        data['old_close'] = data['close']
        for col in ['open', 'high', 'low', 'close']:
            data[col] = (data['r'] * (data[col] * 10 - data['h']) / 10).round(3)
        if 'volume' in data.columns:
            vol = 'volume'
        else:
            vol = 'vol'
        data[vol] = (data['old_close'] * data[vol] / data['close']).round(3)

        if verbose:
            print('after qfq, data:\n', data.loc[rows_to_watch, ['close', 'volume', 'h', 'r']])
    res = data.query('if_trade==True').drop(
        ['if_trade',
         'gubenadj',
         'old_close',
         'h',
         'r',
         'category'],
        axis=1,
        errors='ignore'
    )
    return res


def do_qfq_guben(data, xdxr_data):
    # 股本送配之后/送配前的比例
    xdxr_data = xdxr_data.query('category!=6')
    xdxr_data['gubenadj'] = (10 + xdxr_data['peigu'] + xdxr_data['songzhuangu']) / 10
    xdxr_data['gubenadj'].fillna(1, inplace=True)
    d = xdxr_data[['houzongguben', 'panhouliutong', 'gubenadj']]
    guben = {'zgb': [], 'lgb': []}
    i = 0
    for r in d.itertuples():
        if pd.isnull(r.houzongguben):
            if i == 0: # The first one is just zero
                guben['zgb'].append(0)
                guben['lgb'].append(0)
            else:
                guben['zgb'].append(guben['zgb'][i - 1] * r.gubenadj)
                guben['lgb'].append(guben['lgb'][i - 1] * r.gubenadj)
        else:
            guben['zgb'].append(r.houzongguben)
            guben['lgb'].append(r.panhouliutong)
        i = i + 1
    nv = pd.DataFrame(guben, index=xdxr_data.index)
    xdxr_data = xdxr_data.join(nv)

    mv = xdxr_data.loc[:data.index[-1],
                                       ['zgb',
                                        'lgb']].drop_duplicates()
    data = pd.concat([data, mv], axis=1)
    data[['zgb', 'lgb']] = data[['zgb', 'lgb']].fillna(method='ffill')
    # 市值计算，股本使用复权的数据, 股价不复权, 单位万元。
    data = data.assign(zsz=data.close*data.zgb, lsz=data.close*data.lgb)
    return data


def fq_info(hong, pei, peij, song, h, ratio):
    hong = str(hong)
    pei = str(pei)
    peij = str(peij)
    song = str(song)
    # h = str(h)
    # ratio = str(ratio)
    count = Decimal(10) / (Decimal(10) + Decimal(pei) + Decimal(song))

    fenhong = Decimal(h)/Decimal(count)

    fenhong += Decimal(hong) - Decimal(pei) * Decimal(peij)

    count =  Decimal(count) * Decimal(ratio)
    return (fenhong, count)


def cummulate_rights(xdxr_data):
    idf = xdxr_data[::-1]
    hh = [0]
    rr = [1]
    for r in idf.itertuples():
        lh = hh[-1]
        lr = rr[-1]
        nv = fq_info(r.fenhong, r.peigu, r.peigujia, r.songzhuangu, lh, lr)
        hh.append(nv[0])
        rr.append(nv[1])

    # cumulative fenhong and guben ratio
    xdxr_data['h'] = hh[-1:0:-1]
    xdxr_data['r'] = rr[-1:0:-1]






import gcandle.utils.date_time_utils as date_time_utils
import gcandle.utils.multi_process_pool as executor
from gcandle.objects.basic_objects import DB_CLIENT
from gcandle.objects.data_service_objects import SECURITY_DATA_READ_SERVICE


FP_FEATURE_REPO_NAME_PREFIX = 'fp_features_'
fpds = [20, 50, 120]
MIN_CREATE_DATE = '2006-01-01'


class SlaveIndicatorService:
    def __init__(self, master_service=None):
        self.master_service = master_service

    def set_master(self, master):
        self.master_service = master
        return self

    def save(self, data):
        if data is not None and len(data) > 0:
            repo_name = self.repo_name()
            df = data.drop(columns=['open', 'low', 'high', 'amount', 'vol'
                                  ], axis=1)
            DB_CLIENT.update_data_append_newer(df, repo_name)

    def repo_name(self):
        # not implemented. should be provided by concrete slave indicator service
        return None

    def days_to_update(self):
        # not implemented. should be provided by concrete slave indicator service
        return 0

    def compute_indicator(self, data):
        # not implemented. should be provided by concrete slave indicator service
        return None

    def update_for_single_code(self, code, start, end):
        data = SECURITY_DATA_READ_SERVICE.read_qfq_security_data_for_single(code, start, end)
        if data is None or data.shape[0] < 10:
            return

        master_data = self.master_service.read_master([code], start, end)
        if master_data is None or len(master_data) < 1:
            return
        data = self.compute_indicator(data)
        if data is None or len(data) < 1:
            return
        master_key = self.master_service.get_key()
        data[master_key] = master_data[master_key]
        data = data.loc[data[master_key] == True]
        data.drop(columns=[master_key], inplace=True)
        if data is not None and data.shape[0] > 0:
            self.save(data)

    def update_all_codes(self, codes):
        end = date_time_utils.Date()
        start = end.get_before(days=self.days_to_update())
        executor.execute_tasks(
            codes,
            self.update_for_single_code,
            start.as_str(),
            end.as_str()
        )

    def recreate_all_codes(self, codes, create_start_date, end):
        DB_CLIENT.remove_all_from(self.repo_name())
        DB_CLIENT.create_index_for_collection_with_date_and_code(self.repo_name())
        start = date_time_utils.Date.from_str(create_start_date)
        end = date_time_utils.Date.from_str(end)
        executor.execute_tasks(
            codes,
            self.update_for_single_code,
            start.as_str(),
            end.as_str()
        )

    def read_by_dates(
            self,
            codes,
            start='all',
            end=None
    ):
        repo_name = self.repo_name()
        return DB_CLIENT.read_data_by_codes_and_date(repo_name, codes, start, end=end)


if __name__ == '__main__':
    print("Not callable, abstract class: slave indicator service")


from enum import Enum


class SeFreq(Enum):
    ONE_MIN = '1min'
    FIVE_MIN = '5min'
    FIFTEEN_MIN = '15min'
    THIRTY_MIN = '30min'
    DAY = 'day'
    WEEK = 'week'
    MONTH = 'month'
    YEAR = 'year'

    def is_minutes(self):
        return self == SeFreq.ONE_MIN or \
               self == SeFreq.FIVE_MIN or \
               self == SeFreq.FIFTEEN_MIN or \
               self == SeFreq.THIRTY_MIN

    def keys_to_index(self):
        keys = ['code', 'update_time']
        if self.is_minutes():
            keys.extend(['datetime', 'date'])
        else:
            keys.append('date')
        return keys

    def upper_freq_for_chan(self):
        if self == SeFreq.ONE_MIN:
            return SeFreq.FIVE_MIN
        if self == SeFreq.FIVE_MIN:
            return SeFreq.THIRTY_MIN
        if self == SeFreq.THIRTY_MIN:
            return SeFreq.DAY
        return None

    def lower_freq_for_chan(self):
        if self == SeFreq.MONTH:
            return SeFreq.WEEK
        if self == SeFreq.MONTH:
            return SeFreq.DAY
        if self == SeFreq.DAY:
            return SeFreq.THIRTY_MIN
        if self == SeFreq.THIRTY_MIN:
            return SeFreq.FIVE_MIN
        if self == SeFreq.FIVE_MIN:
            return SeFreq.ONE_MIN
        return None


class SeType(Enum):
    Stock = 'stock'
    Index = 'index'
    ETF = 'etf'

    def prefix_key(self):
        return self.value + '_prefix'

    def repo_name(self, freq: SeFreq):
        return self.value + '_' + freq.value


class Exchange(Enum):
    SZSE = {
        SeType.Stock.prefix_key(): ['00', '30', '02'],
        SeType.Index.prefix_key(): ['39'],
        SeType.ETF.prefix_key(): ['15']
    }
    SHSE = {
        SeType.Stock.prefix_key(): ['6'],
        SeType.Index.prefix_key(): ['000'],
        SeType.ETF.prefix_key(): ['51']
    }

    def contains_code(self, typ:SeType, code):
        for prefix in self.value[typ.prefix_key()]:
            if code.startswith(prefix):
                return True
        return False

    def code_to_type(self, code):
        for t in SeType:
            if self.contains_code(t, code):
                return t
        return None

    def code_to_type_str(self, code):
        t = self.code_to_type(code)
        if t is None:
            return None
        else:
            return t.value

    @classmethod
    def from_type_and_code(cls, typ: SeType, code):
        if Exchange.SHSE.contains_code(typ, code):
            return Exchange.SHSE
        elif Exchange.SZSE.contains_code(typ, code):
            return Exchange.SZSE
        else:
            return None




from gcandle.data.se_types import SeFreq, SeType

class SecurityFetcher:
    def __init__(self):
        pass

    def fetch_bars(self, typ: SeType, code: str, start: str, end: str, freq: SeFreq):
        pass

    def fetch_xdxr(self, code):
        pass

    def fetch_security_list(self):
        pass



import pandas as pd
from gcandle.core.indicator_service.master_indicator_db import MasterIndicatorDB
from gcandle.core.indicator_service.slave_indicator_service import SlaveIndicatorService
from gcandle.utils.date_time_utils import Date

DEFAULT_START_CREATE_DATE = '2006-01-01'


class MasterIndicatorService:
    def __init__(self, name, indicator_func):
        self.name = name
        self.indicator_func = indicator_func
        self.repo = MasterIndicatorDB(self.name, self.indicator_func)
        self.slave_services = []

    def get_key(self):
        return self.name

    def add_slave_service(self, service: SlaveIndicatorService):
        service.set_master(self)
        self.slave_services.append(service)

    def read_all_codes(self):
        return self.repo.read_all_codes()

    def read_master(self, codes, start, end):
        return self.repo.read_for_backtest(codes, start, end)

    def read_master_for_trade(self):
        return self.repo.read_for_trade()

    def read_with_slave_for_backtest(self, codes, start, end, train=None):
        code_list = codes
        if codes is None:
            code_list = self.repo.read_all_codes()
        # code_list = ['600587']
        if train is None:
            master_data = self.repo.read_for_backtest(code_list, start, end)
        elif train:
            master_data = self.repo.read_train_for_backtest(code_list, start, end)
        else:
            master_data = self.repo.read_test_for_backtest(code_list, start, end)

        print('{} loaded'.format(self.name))
        data = self.load_slave_and_join(master_data)
        return data

    def read_with_slave_for_trade(self):
        master_data = self.repo.read_for_trade()
        data = self.load_slave_and_join(master_data)
        return data

    def load_slave_and_join(self, master_data):
        codes = master_data.index.levels[1].unique().tolist()
        start = master_data.index.levels[0].min()
        end = master_data.index.levels[0].max()
        slave_data = self.read_slaves(codes, start, end)
        data = pd.concat([master_data] + slave_data, axis=1, join='outer', sort=True)
        data = data.sort_index(level=0)
        print('Joined')
        return data

    def read_slaves(self, codes, start, end):
        slave_data_list = []
        for service in self.slave_services:
            slave_data_list.append(service.read_by_dates(codes, start, end))
        print('slave data loaded')
        return slave_data_list

    def update_for_all_codes(self):
        self.repo.update_all_codes()
        codes = self.repo.read_all_codes()
        for service in self.slave_services:
            service.update_all_codes(codes)

    def recreate_for_all_codes(self, start=DEFAULT_START_CREATE_DATE, end=None):
        if self.user_confirm_to_continue():
            if end is None:
                end = Date().as_str()
            self.repo.recreate_all_codes(start, end)
            print("{} master indicator recreate done".format(self.name))
            codes = self.repo.read_all_codes()
            for service in self.slave_services:
                service.recreate_all_codes(codes, start, end)
                print("{} slave indicator recreate done".format(service.repo_name()))
        else:
            print("Cancelled")

    def user_confirm_to_continue(self):
        answer = input("Do you really want to recreate all indicators? Enter yes or no: ")
        return answer == 'yes'


if __name__ == '__main__':
    print("Not callable, master indicator service")



from gcandle.indicator.fuquan import stock_to_qfq
from gcandle.objects.basic_objects import DB_CLIENT
from gcandle.objects.data_service_objects import SECURITY_DATA_READ_SERVICE
from gcandle.utils.multi_process_pool import execute_tasks
import gcandle.utils.date_time_utils as date_time_utils


DAY_QFQ_REPO = 'stock_day_qfq'
MIN_START_DATE = '2006-01-01'


def copy_stock_day_directly(code, start, end):
    stock_data = SECURITY_DATA_READ_SERVICE.read_security_data_for_single(code, start, end)
    DB_CLIENT.update_data_append_newer(stock_data, DAY_QFQ_REPO)


def calc_qfq_and_update(code, start, end):
    stock_data = SECURITY_DATA_READ_SERVICE.read_security_data_for_single(code, start, end)
    if stock_data is not None and len(stock_data) > 0:
        qfq_data = stock_to_qfq(stock_data)
        DB_CLIENT.update_data_append_newer(qfq_data, DAY_QFQ_REPO)


def re_calc_qfq_and_overwrite(code, start, end):
    start = MIN_START_DATE
    end = date_time_utils.Date().as_str()
    stock_data = SECURITY_DATA_READ_SERVICE.read_security_data_for_single(code, start, end)
    qfq_data = stock_to_qfq(stock_data)
    DB_CLIENT.replace_data_by_codes(qfq_data, DAY_QFQ_REPO)


def update_qfq_for_single_code(code, start, end):
    max_qfq_date = DB_CLIENT.read_max_date_for(code, DAY_QFQ_REPO)
    if max_qfq_date is None:
        calc_qfq_and_update(code, start, end)
    else:
        fq_info = SECURITY_DATA_READ_SERVICE.read_fq_info(code, start, end)
        if fq_info is None:
            copy_stock_day_directly(code, start, end)
        else:
            max_fq_info_date = fq_info.index.levels[0].max()
            if str(max_fq_info_date)[0:10] < max_qfq_date:
                copy_stock_day_directly(code, start, end)
            else:
                print("Recreate needed for {}, max qfq date={}, max fq info date={}".format(code, max_qfq_date, max_fq_info_date))
                re_calc_qfq_and_overwrite(code, start, end)


def update_stock_qfq_data():
    end = date_time_utils.Date()
    start = end.get_before(days=15)
    code_list = SECURITY_DATA_READ_SERVICE.read_security_codes()
    execute_tasks(code_list, update_qfq_for_single_code, start.as_str(), end.as_str())


def recreate_stock_qfq_data():
    start = MIN_START_DATE
    end = date_time_utils.Date().as_str()
    init_qfq_repo()
    print('START recreate QFQ data')
    code_list = SECURITY_DATA_READ_SERVICE.read_security_codes()
    execute_tasks(code_list, update_qfq_for_single_code, start, end)
    print('DONE recreate QFQ data')


def init_qfq_repo():
    DB_CLIENT.create_index_for_collection_with_date_and_code(DAY_QFQ_REPO)


if __name__=="__main__":
    # recreate_stock_qfq_data()
    update_stock_qfq_data()
    # update_qfq_for_single_code('300831', '2006-01-01', date_time_utils.Date().as_str())





from gcandle.core.indicator_service.indicator_service import IndicatorService
from gcandle.data.se_types import SeFreq
from gcandle.core.indicator_service.fp_mark_service import FPMarkService

DAY_FP_MARK_SERVICES = []

fpds = [20, 50, 120]
for fpd in fpds:
    DAY_FP_MARK_SERVICES.append(FPMarkService(fpd, SeFreq.DAY))



import datetime

DAY_FORMAT = '%Y-%m-%d'
DATETIME_FORMAT = '%Y-%m-%dT%H%M%S'

class Date:
    def __init__(self, s=None, obj=None):
        if s is not None:
            self.obj = datetime.datetime.strptime(s, DAY_FORMAT)
        elif obj is not None:
            self.obj = obj
        else:
            self.obj = datetime.datetime.now()

    @classmethod
    def from_str(cls, s, format=DAY_FORMAT):
        obj = datetime.datetime.strptime(s, format)
        return cls(obj=obj)

    @classmethod
    def from_datetime(cls, obj):
        return cls(obj=obj)

    def as_str(self):
        return datetime.datetime.strftime(self.obj, DAY_FORMAT)

    def as_str_with_time(self):
        return datetime.datetime.strftime(self.obj, DATETIME_FORMAT)

    def as_datetime(self):
        return self.obj

    def get_after(self, days=0, hours=0, minutes=0):
        delta = datetime.timedelta(days=days, hours=hours, minutes=minutes)
        return Date.from_datetime(self.obj + delta)

    def get_before(self, days=0, hours=0, minutes=0):
        delta = datetime.timedelta(days=days, hours=hours, minutes=minutes)
        return Date.from_datetime(self.obj - delta)

    def delta_to(self, other):
        if type(other) == str:
            other = Date.from_str(other)
        return self.obj - other.obj

    #Not considering public holidays, use with caution.
    def next_trade_day(self):
        next_day = self.obj
        if next_day.weekday() == 4:
            next_day = next_day + datetime.timedelta(days=3)
        else:
            next_day = next_day + datetime.timedelta(days=1)
        return Date.from_datetime(next_day)



from gcandle.core.indicator_service.indicator_service import IndicatorService
from gcandle.data.se_types import SeFreq
from gcandle.indicator.price_amount_features import (
    price_foot_peak_mark,
)

FP_MARK_REPO_NAME_PREFIX = 'fp_marks_'
MIN_FP_MARK_DATE = '2016-01-01'


class FPMarkService(IndicatorService):
    def __init__(self, fpd, freq: SeFreq):
        name = FP_MARK_REPO_NAME_PREFIX + str(fpd) + '_' + freq.value
        days = fpd * 3
        super().__init__(name, self.calc_fp_marks, freq=freq, days_to_update=days)
        self.fpd = fpd

    def calc_fp_marks(self, data):
        price_foot_peak_mark(data, self.fpd)
        return data

    def pre_save(self, data):
        if data is not None and len(data) > 0:
            data = data.loc[(data['fp_around_' + str(self.fpd)] > 0)].copy()
            cols = ['open', 'close', 'low', 'high', 'vol', 'amount', 'day_idx']
            if 'hour' in data.columns:
                cols.append('hour')
            if 'minute' in data.columns:
                cols.append('minute')
            cols.extend([c for c in data.columns if 'highest_' in c or 'lowest' in c])
            data.drop(columns=cols, inplace=True)
        return data




from gcandle.data.SecurityDataUpdateMonitor import SecurityDataUpdateMonitor
from gcandle.indicator.basic_indicators import MACD
from gcandle.core.indicator_service.indicator_service import IndicatorService
from gcandle.data.se_types import SeFreq
from gcandle.objects.data_service_objects import SECURITY_DATA_READ_SERVICE
from gcandle.utils.date_time_utils import Date


def macd_calc(data):
    CLOSE = data.close
    df = MACD(CLOSE, 12, 26, 9)
    return df.assign(date=data.date)


DAY_MACD_SERVICE = IndicatorService('stock_day_macd', macd_calc, SeFreq.DAY)
THIRTY_MIN_MACD_SERVICE = IndicatorService('stock_30min_macd', macd_calc, SeFreq.THIRTY_MIN)
FIVE_MIN_MACD_SERVICE = IndicatorService('stock_5min_macd', macd_calc, SeFreq.FIVE_MIN)
ONE_MIN_MACD_SERVICE = IndicatorService('stock_1min_macd', macd_calc, SeFreq.ONE_MIN)


def update_day_macd():
    # codes = SECURITY_DATA_READ_SERVICE.read_security_codes()
    codes = ['000001']
    DAY_MACD_SERVICE.update_all_codes(codes)


def update_30min_macd():
    codes = ['000001']
    THIRTY_MIN_MACD_SERVICE.recreate_all_codes(codes, create_start_date='2018-01-01')
    THIRTY_MIN_MACD_SERVICE.update_all_codes(codes)


def update_5min_macd():
    codes = ['000001']
    FIVE_MIN_MACD_SERVICE.recreate_all_codes(codes, create_start_date='2018-01-01')
    FIVE_MIN_MACD_SERVICE.update_all_codes(codes)


def update_1min_macd():
    codes = ['000001']
    ONE_MIN_MACD_SERVICE.recreate_all_codes(codes, create_start_date='2018-01-01')


def update_all_min_macd():
    monitor = SecurityDataUpdateMonitor()
    a = Date()
    monitor.add_monitor_repo("indicator_stock_day_macd")
    monitor.add_monitor_repo("indicator_stock_1min_macd")
    monitor.add_monitor_repo("indicator_stock_5min_macd")
    monitor.add_monitor_repo("indicator_stock_30min_macd")
    monitor.collect_baseline()
    update_1min_macd()
    update_5min_macd()
    update_30min_macd()
    update_day_macd()
    monitor.collect_info_after_update(a)
    b = Date()
    print("time used: {} seconds".format(b.delta_to(a).seconds))
    monitor.report_changes()

if __name__ == '__main__':
    update_all_min_macd()



import setuptools
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gcandle", # Replace with your own username
    version="0.0.7",
    author="ych",
    author_email="yangchanghua@gmail.com",
    description="A library for quant trade",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/yangchanghua/gcandle",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        "pandas",
        "numpy",
        "pymongo",
        "matplotlib",
        "pytdx"
    ],
)



import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import pandas as pd
from gcandle.core.backtest.backtest_accounting import BacktestAccount
from gcandle.data.se_types import SeType
from gcandle.objects.data_service_objects import SECURITY_DATA_READ_SERVICE


class BacktestAnalysor():
    def __init__(
            self,
            account: BacktestAccount,
            benchmark_code='000300',
            benchmark_name='沪深300',
    ):
        self.account = account
        self.benchmark_code = benchmark_code
        self.benchmark_type = benchmark_name
        bm_data = self._read_bm_data(self.benchmark_code)
        assets_df = self.account.get_daily_assets()
        daily_snapshot = pd.concat([bm_data, assets_df], axis=1, join='outer').fillna(method='pad')
        daily_snapshot.to_csv('backtest_daily_assets_snapshot.csv')
        self._daily_assets = daily_snapshot['asset']
        self._daily_cash = daily_snapshot['cash']
        self._daily_se_value = daily_snapshot['se_value']
        self._bm_assets = (daily_snapshot.close / float(daily_snapshot.close.iloc[0])) * float(self._daily_assets[0])
        self.days_of_benchmark = len(bm_data)
        self.init_cash = self.account.init_cash

    @property
    def max_dropback(self):
        return round(
            float(
                max(
                    [
                        (self._daily_assets.iloc[idx] - self._daily_assets.iloc[idx::].min())
                        / self._daily_assets.iloc[idx]
                        for idx in range(len(self._daily_assets))
                    ]
                )
            ),
            2
        )

    @property
    def total_commission(self):
        return float(
            -abs(round(self.account.history_table.commission.sum(),
                       2))
        )

    @property
    def total_tax(self):
        return float(-abs(round(self.account.history_table.tax.sum(), 2)))

    @property
    def profit_money(self):
        return float(round(self._daily_assets.iloc[-1] - self._daily_assets.iloc[0], 2))

    @property
    def profit(self):
        return round(float(self.calc_profit(self._daily_assets)), 2)

    @property
    def profit_pct(self):
        return self.calc_profitpctchange(self._daily_assets)

    @property
    def annualize_return(self):
        return round(
            float(self.calc_annualize_return(self._daily_assets,
                                             self.days_of_benchmark)),
            2
        )

    def _read_bm_data(self, bm_code):
        return SECURITY_DATA_READ_SERVICE.read_security_data_for_single(
            bm_code,
            self.account.start_date,
            self.account.end_date,
            typ=SeType.Index
        ).reset_index().\
            drop(columns=['code', 'open', 'high', 'low', 'vol', 'amount',
                          'up_count', 'down_count']).\
            set_index('date')

    @property
    def benchmark_profit(self):
        return round(float(self.calc_profit(self._bm_assets)), 2)

    @property
    def benchmark_annualize_return(self):
        return round(
            float(
                self.calc_annualize_return(
                    self._bm_assets,
                    self.days_of_benchmark
                )
            ),
            2
        )

    @property
    def benchmark_profitpct(self):
        return self.calc_profitpctchange(self._bm_assets)

    def set_benchmark(self, code, market_type):
        self.benchmark_code = code
        self.benchmark_type = market_type

    def calc_annualize_return(self, assets, days):
        return round(
            pow(float(assets.iloc[-1]) / float(assets.iloc[0]),
            250 / (float(days))) - 1,
            2
        )

    def calc_profitpctchange(self, assets):
        return assets[::-1].pct_change()[::-1]

    def calc_profit(self, assets):
        return (float(assets.iloc[-1]) / float(assets.iloc[0])) - 1

    def calc_sharpe(self, annualized_returns, volatility_year, r=0.05):
        if volatility_year == 0:
            return 0
        return (annualized_returns - r) / volatility_year

    def plot_assets_curve(self, length=16, height=12):
        print("Ploting assets curve")
        plt.style.use('ggplot')
        plt.figure(figsize=(length, height))
        plt.subplot(211)
        plt.title('Assets/Benchmark')
        self._daily_assets.plot()
        self._bm_assets.plot()
        asset_p = mpatches.Patch(
            color='red',
            label='{}'.format('account')
        )
        asset_b = mpatches.Patch(
            label='benchmark {}'.format(self.benchmark_code)
        )
        plt.legend(handles=[asset_p, asset_b], loc=0)

        plt.subplot(212)
        plt.axis('off')
        xtext = 0.05
        summary = self.summary()
        self.plot_indicator_text(xtext, summary, length)
        return plt

    def plot_indicator_text(self, x, summary, length):
        x_step = 0.2
        y_step = 0.2
        y = 0.8
        items = summary.keys()
        for item in items:
            if x > 0.8:
                x = 0.05
                y -= y_step
            plt.text(
                x,
                y,
                '{} : {}'.format(item,
                                 summary[item]),
                fontsize=12,
                ha='left',
                rotation=0,
                wrap=True
            )
            x += x_step

    def summary(self):
        return {
            'profit rate': round(self.profit, 2),
            'annual profit': round(self.annualize_return, 2),
            'max drop': self.max_dropback,
            'trade days': self.days_of_benchmark,
            'bm annual profit': self.benchmark_annualize_return,
            'initial assets': "%0.2f" % (float(self._daily_assets[0])),
            'final assets': "%0.2f" % (float(self._daily_assets.iloc[-1])),
            'total tax': self.total_tax,
            'total commission': self.total_commission,
        }



import json

import pandas as pd
import pymongo
import gcandle.utils.date_time_utils as date_time_utils
from gcandle.utils.gcandle_config import GCANDLE_CONFIG

READ_BATCH_SIZE = 100000
DEFAULT_PRECISION = 5



def to_json(data):
    if 'date' in data.columns:
        data.date = data.date.apply(lambda x: str(x)[0:10])
    if 'datetime' in data.columns:
        data.datetime = data.datetime.apply(lambda x: str(x)[0:19])
    json_data = json.loads(data.to_json(orient='records'))
    return json_data


class DatabaseClient:
    def __init__(self):
        self.client: pymongo.MongoClient = None

    def set_real_client(self, client):
        self.client = client

    def get_mongo_connection(self):
        return pymongo.MongoClient(GCANDLE_CONFIG.mongodb_uri)

    def get_database(self):
        return self.get_mongo_connection()[GCANDLE_CONFIG.mongodb_name]

    # Multiprocess init can call this function so that every process have their own mongodb connection.
    def init(self):
        self.client = self.get_database()
        return self

    def ensure_client(self):
        if self.client is None:
            self.init()

    def create_index(self, repo_name, keys_list):
        self.ensure_client()
        coll = self.client[repo_name]
        if len(keys_list) > 0:
            for k in keys_list:
                coll.create_index([(k, pymongo.DESCENDING)])
        else:
            print('No keys provided for creating index, NOOP')

    def create_index_for_update_time(self, repo_name):
        self.ensure_client()
        coll = self.client[repo_name]
        coll.create_index(
            [
                ('update_time', pymongo.DESCENDING)
            ]
        )

    def create_index_for_collection_with_date_and_code(self, repo_name):
        self.ensure_client()
        coll = self.client[repo_name]
        coll.create_index(
            [
                ('code', pymongo.ASCENDING),
            ]
        )
        coll.create_index(
            [
                ('date', pymongo.DESCENDING)
            ]
        )
        coll.create_index(
            [
                ('update_time', pymongo.DESCENDING)
            ]
        )
        coll.create_index(
        [
            ('code', pymongo.ASCENDING),
            ('date', pymongo.DESCENDING)
        ], unique=True
        )

    def count_of(self, repo_name, query = None):
        self.ensure_client()
        if query is not None:
            cnt = self.client[repo_name].count_documents(query)
        else:
            cnt = self.client[repo_name].count_documents({})
        return cnt

    def replace_data_with_its_date_range(self, data, repo_name, precision=DEFAULT_PRECISION):
        self.ensure_client()
        if data is None or len(data) < 1:
            return
        data = data.round(precision).reset_index()
        start = data.date.min()
        end = data.date.max()
        code = data.code[0]
        self.delete_by_code_and_dates(repo_name, code, start, end)
        self._insert_data(data, repo_name)

    def replace_data_by_date_range(self, data, repo_name, start, end, precision=DEFAULT_PRECISION):
        self.ensure_client()
        if data is None or len(data) < 1:
            return
        data = data.round(precision).reset_index()
        code = data.code[0]
        self.delete_by_code_and_dates(repo_name, code, start, end)
        self._insert_data(data, repo_name)

    def delete_by_code_and_dates(self, repo_name, code, start, end):
        coll = self.client[repo_name]
        coll.delete_many({'code': {'$in': [code]}, 'date': {
            '$gte': start, '$lte': end
        }})

    def _insert_data(self, data, repo_name):
        coll = self.client[repo_name]
        data['update_time'] = date_time_utils.Date().as_str_with_time()
        return coll.insert_many(to_json(data))

    def replace_data_by_codes(self, data, repo_name, precision=DEFAULT_PRECISION):
        self.ensure_client()
        if data is None or len(data) < 1:
            return
        data = data.round(precision).reset_index()
        codes = data.code.unique().tolist()
        coll = self.client[repo_name]
        coll.delete_many({'code': {'$in': codes}})
        if len(data) != 0:
            data['update_time'] = date_time_utils.Date().as_str_with_time()
            return coll.insert_many(to_json(data))

    def update_data_append_newer(self, data, repo_name, precision=DEFAULT_PRECISION):
        self.ensure_client()
        if data is None or len(data) < 1:
            return
        data = data.round(precision).reset_index()
        codes = data.code.unique()
        keep_flag = 'newer_to_keep'
        data[keep_flag] = False
        for code in codes:
            max_date = self.read_max_date_for(code, repo_name)
            code_filter = data.code == code
            if max_date is None:
                data.loc[code_filter, keep_flag] = True
            elif len(data.loc[code_filter & (data.date > max_date)]) > 0:
                data.loc[code_filter & (data.date > max_date), keep_flag] = True
        data = data.loc[data[keep_flag] == True]
        if data is not None and len(data) > 0:
            data['update_time'] = date_time_utils.Date().as_str_with_time()
            data.drop(columns=[keep_flag], inplace=True)
            self.client[repo_name].insert_many(to_json(data))

    def read_codes_from_db(self, repo_name, filter=None):
        self.ensure_client()
        db = self.client[repo_name]
        res = db.distinct('code', filter=filter)
        return res

    def construct_dataframe(self, cursor):
        self.ensure_client()
        res = pd.DataFrame([item for item in cursor])
        if res.shape[0] < 1:
            return None
        try:
            if 'datetime' in res.columns:
                time_index = 'datetime'
            else:
                time_index = 'date'
            res = res.assign(date=pd.to_datetime(
                res.date)).drop_duplicates(([time_index, 'code']))
            res['code'] = res['code'].astype(str)
            res = res.set_index([time_index, 'code'], drop=True).sort_index(level=0)
        except:
            print('Cannot convert data from db to dataframe')
            res = None
        return res

    def read_all_data(self, repo_name):
        self.ensure_client()
        coll = self.client[repo_name]
        cursor = coll.find({}, {'_id': 0}, batch_size=READ_BATCH_SIZE)
        return self.construct_dataframe(cursor)

    def read_data_by_codes_and_date(
            self,
            repo_name,
            codes,
            start, end=None,
            extra_filter=None,
            ):
        if end is None:
            end = date_time_utils.Date().as_str()
        else:
            end = str(end)[0:10]
        start = str(start)[0:10]

        query = {
            'code': {'$in': codes},
            "date": {
                "$lte": end,
                "$gte": start
            }
        }
        if extra_filter is not None:
            query.update(extra_filter)
        coll = self.client[repo_name]
        cursor = coll.find(query, {"_id": 0, 'update_time': 0, 'date_stamp': 0}, batch_size=READ_BATCH_SIZE)
        return self.construct_dataframe(cursor)

    def read_max_date_for(
            self, code, repo_name
            ):
        self.ensure_client()
        repo = self.client[repo_name]
        cnt = repo.count_documents({'code': code})
        if cnt < 1:
            return None
        end = repo.find({'code': code}, {'date': 1, '_id': 0}).sort('date', pymongo.DESCENDING).limit(1)[0]['date']
        return end[0:10]

    def remove_all_from(self, repo_name, query=None):
        self.ensure_client()
        repo = self.client[repo_name]
        if query is None:
            repo.drop()
        else:
            repo.remove(query)

    def drop(self, repo_name):
        self.ensure_client()
        repo = self.client[repo_name]
        repo.drop()




import logging
import time
import numpy as np
import pandas as pd
from gcandle.core.backtest.backtest_accounting import BacktestAccount
from gcandle.core.backtest_analysis import BacktestAnalysor
from gcandle.core.sell_strategy.AbstractSellStrategy import (AbstractSellStrategy, SPR_KEY, SELL_STRATEGY_KEY)
from gcandle.core.sell_strategy.CombinedSellStrategy import CombinedSellStrategy
from gcandle.core.buy_strategy.BuyStrategy import BuyStrategy
from gcandle.core.common_types import ORDER_DIRECTION

DEFAULT_CASH = 100000


class DayBarBacktest:
    def __init__(self, name):
        self.name = name
        self.data = None
        self.init_cash = DEFAULT_CASH
        self.buy_strategy: BuyStrategy = None
        self.sell_strategy: CombinedSellStrategy = None

    def set_data(self, data):
        self.data = data
        return self

    def set_init_cash(self, cash):
        self.init_cash = cash
        return self

    def set_buy_strategy(self, strategy: BuyStrategy):
        self.buy_strategy = strategy
        return self

    def set_sell_strategy(self, strategy: AbstractSellStrategy):
        self.sell_strategy = strategy
        return self

    def init_account(self, init_cash):
        AC = BacktestAccount(init_cash=init_cash)
        self.AC = AC
        self.orders = []

    def choose_cand_without_limit(self, buy_cand):
        max_ratio = 1.0
        chosen_buy = self.choose_cand(buy_cand)
        return chosen_buy, max_ratio

    def choose_cand_with_max_ratio(self, buy_cand, hold_cnt):
        can_buy_cnt = max(1, 3 - hold_cnt)
        if hold_cnt == 0:
            max_ratio = 0.33
        elif hold_cnt == 1:
            max_ratio = 0.5
        else:
            max_ratio = 1
        chosen_buy = self.choose_cand(buy_cand, can_buy_cnt)
        return chosen_buy, max_ratio

    def print_sell_stock(self, item):
        arr = self.get_item_attrs(item)
        print('SELL ', arr)

    def get_item_attrs(self, item):
        arr = []
        arr.append(item.date)
        arr.append(item.code)
        arr.append(item.xian_a)
        arr.append(item.xianBC)
        arr.append(item.pLAXChg)
        arr.append(item.pHAXChg)
        arr.append(item.low)
        arr.append(item.close)

    def output_order_stats(self):
        tt = pd.DataFrame(self.orders, columns=['date', 'code', 'blow', 'bpr', 'spr', 'shigh', 'sdate'])
        tt['profit'] = tt.spr / tt.bpr
        tt['win'] = tt.profit >= 1.03
        tt['loss'] = (tt.win == False) & (tt.profit <= 0.97)
        tt['avg'] = (tt.win == False) & (tt.loss == False)
        date_filter = (tt.date > '2000-10-01') & (tt.date < '2020-12-30')
        profit_filter = tt.profit < 0.9
        df = tt.loc[date_filter & profit_filter]
        win_cnt = len(tt.loc[tt.win])
        loss_cnt = len(tt.loc[tt.loss])
        win_ratio = win_cnt / len(tt)
        loss_ratio = loss_cnt / len(tt)
        print(
            'win, loss cnt {}, {}, ratio: {}, {}'.format(win_cnt, loss_cnt, round(win_ratio, 2), round(loss_ratio, 2)))

    def print_buy_stock(self, item):
        arr = self.get_item_attrs(item)
        print('BUY ', arr)

    def choose_cand(self, cand, max_cnt=30):
        cnt = len(cand)
        chosen = []
        if cnt < 1:
            chosen = []
        elif cnt <= max_cnt:
            chosen = [c for c in cand]
        else:
            n = np.random.randint(cnt - max_cnt)
            chosen = [cand[i] for i in range(n, n + max_cnt)]
        return chosen

    def buy(self, code, item, ratio, price_func):
        self.AC.order(data=item, towards=ORDER_DIRECTION.BUY, ratio=ratio, \
                      price_func=price_func)

    def sell(self, code, item, price):
        self.AC.order(data=item, towards=ORDER_DIRECTION.SELL, price=price, ratio=1.0)

    def get_active_position_days(self, data):
        return self.AC.get_active_position_days(data.code)

    def sell_immediately(self, sell_data):
        data = sell_data.copy()
        today = data['date'][0]
        today = str(today)[0:10]
        # print('sell on {}'.format(today))
        sellable = self.AC.sell_available
        if len(sellable) < 1:
            return
        data = data.loc[data.code.isin(sellable)].copy()
        if len(data) <= 0:
            logging.debug("{} No data for sell, sellable = {}".format(today, sellable))
            return
        hold_days = data.apply(self.get_active_position_days, axis=1)
        data['hold_days'] = hold_days
        spr = self.sell_strategy.get_prices(data)
        cols = [SPR_KEY, SELL_STRATEGY_KEY]
        data[cols] = spr[cols]
        for c in sellable:
            if c not in data['code'].values:  # fixed bug: 股票当日停牌,无法交易
                # print('{} No data for today, maybe paused for trading'.format(c))
                pass
            else:
                item = data.loc[data.code == c].squeeze()
                price = item[SPR_KEY]
                name = item[SELL_STRATEGY_KEY]
                if price > 0:
                    self.sell(item.code, item, price=price)
                else:
                    pass

    def onbar(self, data):
        data = data.reset_index().copy()
        today = str(data.date[0])[0:10]
        if today == '2016-11-25':
            print(today)
        AC = self.AC

        f = self.buy_strategy.get_final_buy_filter_on_buy_day(data)
        f &= ~data.code.isin(self.AC.sell_available)
        buy_data = data.loc[f]

        if self.sell_first:
            self.sell_immediately(data)

        if len(buy_data) > 0:
            buy_cand_codes = buy_data.code.unique().tolist()
            chosen_buy_codes, max_ratio = self.choose_cand_without_limit(buy_cand_codes)
            buy_data = buy_data.loc[buy_data.code.isin(chosen_buy_codes)].copy()

            price_func = self.buy_strategy.get_price_func()
            self.AC.batch_buy(buy_data, price_func, max_ratio=max_ratio)
            for c in chosen_buy_codes:
                item = buy_data.loc[buy_data['code'] == c].squeeze()

        if not self.sell_first:
            self.sell_immediately(data)

        AC.settle(data)

    def run(self, y, yy, m=1, mm=12, sell_first=False, extra_filter=False, silent=False):
        if not self._ready_to_run():
            print('Exit now.')
            return
        bg = time.time()
        self.init_account(self.init_cash)
        self.sell_first = sell_first
        self.use_extra_filter = extra_filter

        data = self.data.copy()
        # data = data.loc[datetime(y, m, 1):datetime(yy, mm, 30)]

        data = self.buy_strategy.calc_and_forward_fill_buy_filter(data, extra=self.use_extra_filter)

        self.buy_strategy.update_buy_prices(data)
        tm = time.time()
        print('Time used for calc buy and price: ', round(tm - bg, 2))
        AC = self.AC
        data = data.sort_index(level=0)
        data.groupby(level=0).apply(self.onbar)

        if not silent:
            print('Time used for on bar: ', time.time() - tm)
            print('Win rate: ', AC.win_rate())
            risk = BacktestAnalysor(AC)
            fig = risk.plot_assets_curve()
            fig.show()

        AC.market_data = data
        his_table = AC.history_table

        hist_file = open('{}_history.html'.format(self.name), 'w')
        print(his_table.to_html(), file=hist_file)
        hist_file.close()

        cols = ['datetime', 'code', 'price', 'amount', 'direction']

        if not silent:
            print('### Total orders ###: ', his_table.shape[0] / 2)
            print(self.sell_strategy.get_history())

        top = AC.max_profitable_stocks(n=20)
        bottom = AC.max_profitable_stocks(n=-20)
        f = open('{}_top_profit.html'.format(self.name), 'w')
        print(top.T.to_html(), file=f)
        top_codes = list(top.index.values)
        print('Top profit codes: ', top_codes)
        top_hist = his_table.loc[his_table['code'].isin(top_codes), cols]
        print(top_hist.to_html(), file=f)
        f.close()

        f = open('{}_bottom_profit.html'.format(self.name), 'w')
        print(bottom.to_html(), file=f)
        bottom_codes = list(bottom.index.values)
        print('Bottom profit codes: ', bottom_codes)
        bottom_hist = his_table.loc[his_table['code'].isin(bottom_codes), cols]
        print(bottom_hist.to_html(), file=f)
        f.close()

    def _ready_to_run(self):
        if self.data is None:
            print('No data is set for backtest')
            return False
        if self.buy_strategy is None:
            print('Buy strategy is not set yet.')
            return False
        if self.sell_strategy is None:
            print('Sell strategy is not set yet.')
            return False
        return True



from concurrent.futures import ThreadPoolExecutor
import gcandle.utils.date_time_utils as date_time_utils
from gcandle.objects.basic_objects import DB_CLIENT
import gcandle.utils.multi_process_pool as executor
from gcandle.data.SecurityFetcher import SecurityFetcher
from gcandle.data.se_types import SeFreq, SeType
from gcandle.utils.CodeListTaskExecutor import CodeListTaskExecutor

MIN_START_TIME = '2006-01-01'
REPO_XDXR = 'security_xdxr'
REPO_SECURITY_LIST = 'security_list'
REPO_QFQ_SUFFIX = '_qfq'
DAYS_TO_UPDATE = 16


class SecurityDataService:
    def __init__(self, fetcher: SecurityFetcher = None):
        self.fetcher = fetcher
        self.failed_codes = []

    def set_fetcher(self, fetcher: SecurityFetcher):
        self.fetcher = fetcher

    def read_security_codes(self, typ: SeType = SeType.Stock):
        return DB_CLIENT.read_codes_from_db(REPO_SECURITY_LIST, filter={'stype': typ.value})

    def read_fq_info(self, code, start, end):
        data = DB_CLIENT.read_data_by_codes_and_date(REPO_XDXR, [code], start, end=end, extra_filter={'category': 1})
        return data

    def read_qfq_security_data_for_single(self, code, start, end, typ: SeType = SeType.Stock, freq: SeFreq = SeFreq.DAY):
        codes = [code]
        data = DB_CLIENT.read_data_by_codes_and_date(typ.repo_name(freq) + REPO_QFQ_SUFFIX, codes, start, end)
        if data is None:
            return None
        return data[data.vol > 1]

    def read_security_data_for_single(self, code, start, end, typ: SeType = SeType.Stock, freq: SeFreq = SeFreq.DAY):
        codes = [code]
        data = DB_CLIENT.read_data_by_codes_and_date(typ.repo_name(freq), codes, start, end)
        if data is None:
            return None
        if 'vol' in data.columns:
            return data[data.vol > 1]
        else:
            return data

    def read_security_data_for_list(self, codes, start, end, typ: SeType = SeType.Stock, freq: SeFreq = SeFreq.DAY):
        return DB_CLIENT.read_data_by_codes_and_date(typ.repo_name(freq), codes, start, end)

    def do_fetch_and_save_bars(self, code, typ: SeType, start, end, freq: SeFreq):
        repo_name = typ.repo_name(freq)
        end_date = date_time_utils.Date().as_str()
        max_date_in_db = DB_CLIENT.read_max_date_for(code, repo_name)
        if max_date_in_db is None:
            start_date = MIN_START_TIME
        else:
            start_date = date_time_utils.Date.from_str(max_date_in_db).as_str()
        try:
            if start_date != end_date:
                df = self.fetcher.fetch_bars(typ, code, start, end, freq)
                if len(df) < 1:
                    self.failed_codes.append(code)
                else:
                    df = self.remove_columns_for_min_data(df, freq)
                    DB_CLIENT.update_data_append_newer(df, repo_name)
        except Exception as e:
            print('Failed to fetch bars for {} {}, reason={}'.format(code, freq, e))
            self.failed_codes.append(str(code))

    def do_fetch_and_replace_by_dates(self, code, typ: SeType, start, end, freq: SeFreq):
        repo_name = typ.repo_name(freq)
        df = self.fetcher.fetch_bars(typ, code, start, end, freq)
        if df is None or len(df) < 1:
            self.failed_codes.append(code)
        else:
            df = self.remove_columns_for_min_data(df, freq)
            DB_CLIENT.replace_data_with_its_date_range(df, repo_name)

    def get_failed_codes(self):
        return self.failed_codes

    def update_bars_by_dates(self,  start, end, typ: SeType=SeType.Stock, freq: SeFreq = SeFreq.DAY, codes=None):
        DB_CLIENT.create_index(typ.repo_name(freq), freq.keys_to_index())
        if codes is None:
            codes = self.read_security_codes(typ)
        executor.execute_tasks(codes, self.do_fetch_and_replace_by_dates, typ, start, end, freq)
        print("These codes failed to update: [" + ','.join(self.failed_codes) + "]")

    def update_bars(self, typ: SeType=SeType.Stock, freq: SeFreq = SeFreq.DAY, codes=None, days_to_update=DAYS_TO_UPDATE):
        DB_CLIENT.create_index(typ.repo_name(freq), freq.keys_to_index())
        if codes is None:
            codes = self.read_security_codes(typ)
        end = date_time_utils.Date()
        start = end.get_before(days_to_update)
        executor.execute_tasks(codes, self.do_fetch_and_save_bars, typ, start.as_str(), end.as_str(), freq)
        print("These codes failed to update: [" + ','.join(self.failed_codes) + "]")

    def refetch_and_save_bars(self, typ: SeType, freq: SeFreq = SeFreq.DAY, codes=None):
        if codes is None:
            if self.__confirm_drop():
                self.drop_bar_repo(typ, freq)
                print('删除完成。')
                codes = self.read_security_codes(typ)
            else:
                print("操作已取消。")
                return
        else:
            DB_CLIENT.remove_all_from(typ.repo_name(freq), query={'code': {'$in': codes}})
        DB_CLIENT.create_index(typ.repo_name(freq), freq.keys_to_index())
        end = date_time_utils.Date()
        start = date_time_utils.Date.from_str(MIN_START_TIME)
        executor.execute_tasks(codes, self.do_fetch_and_save_bars, typ, start.as_str(), end.as_str(), freq)

    def drop_bar_repo(self, typ: SeType, freq: SeFreq):
        DB_CLIENT.drop(typ.repo_name(freq))

    def __confirm_drop(self):
        confirm = input("确认删除所有的数据？请确认[yes/no]:")
        return confirm == 'yes'

    def do_update_xdxr(self, code):
        df = self.fetcher.fetch_xdxr(code)
        if len(df) > 0:
            DB_CLIENT.update_data_append_newer(df, REPO_XDXR)

    def refetch_and_save_xdxr(self, codes: list):
        executor = CodeListTaskExecutor(ThreadPoolExecutor)
        executor.execute(codes, self.do_update_xdxr)

    def update_code_list(self):
        data = self.fetcher.fetch_security_list()
        DB_CLIENT.replace_data_by_codes(data, REPO_SECURITY_LIST)

    def remove_columns_for_min_data(self, df, freq):
        if freq == SeFreq.ONE_MIN or freq == SeFreq.FIVE_MIN:
            df = df.drop(columns=['vol', 'amount', 'open', 'hour', 'minute'], errors='ignore')
        return df


STOCK_DAY_DATA_SERVICE = SecurityDataService()


if __name__ == '__main__':
    from gcandle.data.tdx.TdxSecurityFetcher import TdxSecurityFetcher
    service = SecurityDataService(TdxSecurityFetcher())
    # service.update_code_list()
    # service.update_xdxr(codes)
    # service.refetch_and_save_bars(typ=SeType.Stock, freq=SeFreq.FIVE_MIN)
    service.refetch_and_save_bars(typ=SeType.Stock, freq=SeFreq.THIRTY_MIN)
    # service.do_fetch_and_save_bars('000001', SeType.Stock, '2020-08-01', '2020-09-30', SeFreq.FIVE_MIN)






# Interface for buy strategy
class BuyStrategy:
    def __init__(self, name):
        self.name = name
        self._buy_filter_func = None
        self._price_func = None

    def update_buy_prices(self, data):
        pass

    # can be used both for candidate scan and backtest.
    def get_buy_filter_before_buy_day(self, data):
        return True

    def get_final_buy_filter_on_buy_day(self, data):
        if self._buy_filter_func is not None:
            return self._buy_filter_func(data)
        else:
            return False

    def calc_and_forward_fill_buy_filter(self, data, extra=True):
        return data

    def get_price_func(self):
        return self._price_func

    def set_filter_on_buy_day(self, filter_func):
        self._buy_filter_func = filter_func
        return self

    def set_price_func(self, price_func):
        self._price_func = price_func
        return self

